/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.storage.jdbc;

import java.io.IOException;
import java.io.InputStream;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

import javax.jcr.PropertyType;
import javax.sql.DataSource;

import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.storage.value.ValueStoragePluginProvider;

/**
 * Created by The eXo Platform SARL        . </br>
 * 
 * Concrete JDBC based data container that uses "table-set per Workspace policy"
 * i.e each JCR Workspace storage is placed in dedicated DB like:
 * table JCR_MITEM (ID varchar(255) not null, VERSION integer not null, PATH varchar(255) not null, primary key (ID))
 * table JCR_MNODE (ID varchar(255) not null, ORDER_NUM integer, PARENT_ID varchar(255), primary key (ID))
 * table JCR_MPROPERTY (ID varchar(255) not null, TYPE integer not null, PARENT_ID varchar(255) not null, MULTIVALUED bit not null, primary key (ID))
 * table JCR_MVALUE (ID bigint not null, DATA varbinary(65535) not null, ORDER_NUM integer, PROPERTY_ID varchar(255) not null, primary key (ID))
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: MultiDbJDBCConnection.java 8556 2006-09-04 15:32:25Z peterit $
 */

public class MultiDbJDBCConnection extends JDBCStorageConnection {
  
  
  MultiDbJDBCConnection(DataSource dataSource, String containerName, ValueStoragePluginProvider valueStorageProvider) throws SQLException {
    super(dataSource, containerName, true, valueStorageProvider);
  }
  
  /**
   * Prepared queries at start time
   * @throws SQLException
   */
  @Override
  protected final void prepareQueries() throws SQLException {
    JCR_FK_NODE_PARENT = "JCR_FK_MNODE_PARENT";
    JCR_FK_NODE_ITEM = "JCR_FK_MNODE_ITEM";
    JCR_FK_PROPERTY_NODE = "JCR_FK_MPROPERTY_N";
    JCR_FK_PROPERTY_ITEM = "JCR_FK_MPROPERTY_I";
    JCR_FK_VALUE_PROPERTY = "JCR_FK_MVALUE_PROP";
     
    FIND_ITEM_BY_ID = "select I.*, N.ID as NID, N.ORDER_NUM as NORDER_NUM, N.PARENT_ID as NPARENT_ID, P.ID as PID, P.TYPE as PTYPE, P.PARENT_ID as PPARENT_ID, P.MULTIVALUED as PMULTIVALUED" 
      + " from JCR_MITEM I LEFT JOIN JCR_MNODE N ON I.ID=N.ID LEFT JOIN JCR_MPROPERTY P ON I.ID=P.ID"
      + " where I.ID=?";

    FIND_ITEM_BY_PATH = "select I.*, N.ID as NID, N.ORDER_NUM as NORDER_NUM, N.PARENT_ID as NPARENT_ID, P.ID as PID, P.TYPE as PTYPE, P.PARENT_ID as PPARENT_ID, P.MULTIVALUED as PMULTIVALUED" 
      + " from JCR_MITEM I LEFT JOIN JCR_MNODE N ON I.ID=N.ID LEFT JOIN JCR_MPROPERTY P ON I.ID=P.ID"
      + " where I.PATH=?";
    
    FIND_CHILDITEM_BY_PATH = "select I.*, N.ID as NID, N.ORDER_NUM as NORDER_NUM, N.PARENT_ID as NPARENT_ID, P.ID as PID, P.TYPE as PTYPE, P.PARENT_ID as PPARENT_ID, P.MULTIVALUED as PMULTIVALUED" 
      + " from JCR_MITEM I left join JCR_MNODE N on (I.ID=N.ID and N.PARENT_ID=?) left join JCR_MPROPERTY P on (I.ID=P.ID and P.PARENT_ID=?)"
      + " where I.PATH=?";

    FIND_PROPERTIES_BY_VALUE = "select I.*, P.ID as PID, P.TYPE as PTYPE, P.PARENT_ID as PPARENT_ID, P.MULTIVALUED as PMULTIVALUED, V.ORDER_NUM as VORDER_NUM, V.PROPERTY_ID as VPROPERTY_ID, V.DATA as VDATA"   
      + " from JCR_MITEM I, JCR_MPROPERTY P, JCR_MVALUE V"
      + " where I.ID=P.ID and P.ID=V.PROPERTY_ID and P.TYPE=" + PropertyType.REFERENCE
      + " and V.DATA=?"
      + " order by I.ID";
    
    FIND_VALUES_BY_PROPERTYID = "select * from JCR_MVALUE where PROPERTY_ID=? order by ORDER_NUM";
    FIND_VALUE_BY_PROPERTYID_OREDERNUMB = "select * from JCR_MVALUE where PROPERTY_ID=? and ORDER_NUM=?";
    
    FIND_NODES_BY_PARENTID = "select I.*, N.ID as NID, N.ORDER_NUM as NORDER_NUM, N.PARENT_ID as NPARENT_ID" 
      + " from JCR_MITEM I, JCR_MNODE N"
      + " where I.ID=N.ID and N.PARENT_ID=?" 
      + " order by I.ID"; // " order by I.ID DESC" for Node types registration tests
    
    FIND_NODES_IDS_BY_PARENTID = "select I.ID" 
      + " from JCR_MITEM I, JCR_MNODE N"
      + " where I.ID=N.ID and N.PARENT_ID=?" 
      + " group by I.ID"     
      + " order by I.ID";
    
    FIND_PROPERTIES_BY_PARENTID = "select I.*, P.ID as PID, P.TYPE as PTYPE, P.PARENT_ID as PPARENT_ID, P.MULTIVALUED as PMULTIVALUED" 
      + " from JCR_MITEM I, JCR_MPROPERTY P"
      + " where I.ID=P.ID and P.PARENT_ID=?" 
      + " order by I.ID";
    
    FIND_NODESCOUNT_BY_PARENTID = "select count(*) from JCR_MNODE where PARENT_ID=?";
    FIND_PROPERTIESCOUNT_BY_PARENTID = "select count(*) from JCR_MPROPERTY where PARENT_ID=?";
    
    INSERT_ITEM = "insert into JCR_MITEM(ID, PATH, VERSION) VALUES(?,?,?)";
    INSERT_NODE = "insert into JCR_MNODE(ID, ORDER_NUM, PARENT_ID) VALUES(?,?,?)";
    INSERT_PROPERTY = "insert into JCR_MPROPERTY(ID, TYPE, MULTIVALUED, PARENT_ID) VALUES(?,?,?,?)";
    INSERT_VALUE = "insert into JCR_MVALUE(DATA, ORDER_NUM, PROPERTY_ID) VALUES(?,?,?)";

    UPDATE_ITEM = "update JCR_MITEM set VERSION=? where ID=?";
    UPDATE_NODE = "update JCR_MNODE set ORDER_NUM=? where ID=?";
    UPDATE_PROPERTY = "update JCR_MPROPERTY set TYPE=? where ID=?";
    //UPDATE_VALUE = "update JCR_MVALUE set DATA=?, ORDER_NUM=? where ID=?";
    
    DELETE_ITEM = "delete from JCR_MITEM where ID=?";
    DELETE_NODE = "delete from JCR_MNODE where ID=?";
    DELETE_PROPERTY = "delete from JCR_MPROPERTY where ID=?";
    DELETE_VALUE = "delete from JCR_MVALUE where PROPERTY_ID=?";    
  }  

  @Override
  protected void addNodeRecord(NodeData data) throws SQLException {
    insertItem.setString(1, data.getUUID());
    insertItem.setString(2, data.getQPath().getAsString());
    insertItem.setInt(3, data.getPersistedVersion());
    insertItem.executeUpdate();

    insertNode.setString(1, data.getUUID());
    insertNode.setInt(2, data.getOrderNumber());
    insertNode.setString(3, data.getParentUUID());
    insertNode.executeUpdate();    
  }

  @Override
  protected void addPropertyRecord(PropertyData data) throws SQLException {
    insertItem.setString(1, data.getUUID());
    insertItem.setString(2, data.getQPath().getAsString());
    insertItem.setInt(3, data.getPersistedVersion());
    insertItem.executeUpdate();

    insertProperty.setString(1, data.getUUID());
    insertProperty.setInt(2, data.getType());
    insertProperty.setBoolean(3, data.isMultiValued());
    insertProperty.setString(4, data.getParentUUID());
    insertProperty.executeUpdate();
//    addValueRecords(data);
  }
  
  @Override
  protected void addValueRecords(PropertyData data) throws SQLException, IOException {
    List<ValueData> values = data.getValues();
    for(int i=0; i<values.size(); i++) {
      ValueData val = values.get(i);
      // [PN] 17.08.06
      InputStream stream = val.getAsStream();
      insertValue.setBinaryStream(1, stream, stream.available());
      
      // [PN] 26.04.06 Natural values order using
      // [GA] already set in doUpdateProperty()
      //val.setOrderNumber(i);
      insertValue.setInt(2, val.getOrderNumber());
      
      insertValue.setString(3, data.getUUID());
      insertValue.executeUpdate();
      
      // [GA] do we need to close or executeUpdate() does it? 
      stream.close();

    }
  }  

  @Override
  protected int deleteItemByUUID(String uuid) throws SQLException {
    deleteItem.setString(1, uuid);
    return deleteItem.executeUpdate();
  }

  @Override
  protected int deleteNodeByUUID(String uuid) throws SQLException {
    deleteNode.setString(1, uuid);
    return deleteNode.executeUpdate();
  }

  @Override
  protected int deletePropertyByUUID(String uuid) throws SQLException {
    deleteProperty.setString(1, uuid);
    return deleteProperty.executeUpdate();
  }

  @Override
  protected int deleteValueByUUID(String uuid) throws SQLException {
    deleteValue.setString(1, uuid);
    return deleteValue.executeUpdate();
  }

  @Override
  protected int updateItemVersionByUUID(int versionValue, String uuid) throws SQLException {
    updateItem.setInt(1, versionValue);
    updateItem.setString(2, uuid);
    return updateItem.executeUpdate();
  }
  
  @Override
  protected int updateNodeOrderNumbByUUID(int orderNumb, String uuid) throws SQLException {
    updateNode.setInt(1, orderNumb);
    updateNode.setString(2, uuid);
    return updateNode.executeUpdate();
  }
  
  @Override
  protected int updatePropertyTypeByUUID(int type, String uuid) throws SQLException {
    updateProperty.setInt(1, type);
    updateProperty.setString(2, uuid);
    return updateProperty.executeUpdate();
  }
  
  @Override
  protected ResultSet findItemByPath(String path) throws SQLException {
    findItemByPath.setString(1, path);
    
    return findItemByPath.executeQuery();
  }
  
  @Override
  protected ResultSet findItemByPath(String parentId, String path) throws SQLException {
    findChildItemByPath.setString(1, parentId);
    findChildItemByPath.setString(2, parentId);
    findChildItemByPath.setString(3, path);
    return findChildItemByPath.executeQuery();
  }

  @Override
  protected ResultSet findItemByUUID(String uuid) throws SQLException {
    findItemById.setString(1, uuid);
    return findItemById.executeQuery();
  }

  @Override
  protected ResultSet findReferences(String uuid) throws SQLException {
    findPropertiesByValue.setBytes(1, uuid.getBytes());
    return findPropertiesByValue.executeQuery();
  }

  @Override
  protected ResultSet findValuesByPropertyId(String propertyId) throws SQLException {
    findValuesByPropertyId.setString(1, propertyId);
    return findValuesByPropertyId.executeQuery();
  }
  
  @Override
  protected ResultSet findValue(String valueId, int orderNumb) throws SQLException {
    findValue.setString(1, valueId);
    findValue.setInt(2, orderNumb);
    return findValue.executeQuery();
  }

  @Override
  protected ResultSet findChildNodesByParentUUID(String parentUUID) throws SQLException {
    findNodesByParentId.setString(1, parentUUID);
    return findNodesByParentId.executeQuery();
  }
  
  @Override
  protected ResultSet findChildNodesIdsByParentUUID(String parentUUID) throws SQLException {
    findNodesIdsByParentId.setString(1, parentUUID);
    return findNodesIdsByParentId.executeQuery();
  }

  @Override
  protected ResultSet getChildNodesCountByParentUUID(String parentUUID) throws SQLException {
    findNodesCountByParentId.setString(1, parentUUID);
    return findNodesCountByParentId.executeQuery();
  }

  @Override
  protected ResultSet findChildPropertiesByParentUUID(String parentUUID) throws SQLException {
    findPropertiesByParentId.setString(1, parentUUID);
    return findPropertiesByParentId.executeQuery();
  }

  @Override
  protected ResultSet getChildPropertiesCountByParentUUID(String parentUUID) throws SQLException {
    findPropertiesCountByParentId.setString(1, parentUUID);
    return findPropertiesCountByParentId.executeQuery();
  }

}
