/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.datamodel;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.storage.swap.BinaryValueSwapStorage;
import org.exoplatform.services.jcr.util.io.BLOBUtil;
import org.exoplatform.services.jcr.util.io.CloneableInputStream;
import org.exoplatform.services.jcr.util.io.LazyByteArrayInputStream;
import org.exoplatform.services.jcr.util.io.LazyTempFileInputStream;
import org.exoplatform.services.jcr.util.io.SharedFileInputStream;
import org.exoplatform.services.jcr.util.io.TempFile;
import org.exoplatform.services.log.ExoLogger;

/**
 * AbstractValueData class provide common functions for value data storing. 
 * With support of temporary swapping in local files.
 * AbstractValueData can store value data as array and as stream of bytes.
 * 
 * In most of cases data will be stored as array of bytes.
 * In case of BLOB (binary large objects) it may be stored in input stream.
 * 
 * If used 'inBuffer' option then given stream will be copied into bytes array.   
 * Otherwise, a ValueData will store given InputStream as direct source of data access in
 * getAsStream() method. 
 * Except of FileInputStream and CloneableInputStream instances which will be cloned.
 * 
 * If we will use local swapping inside value data object we have to use constructors with 'swapStorage' parameter.
 * But swapping will be available for BLOB (binary large objects) values only.
 * Data will be detected as BLOB if data has size (bytes count) bigger given limit of the swap storage.
 * 
 * If swap is enabled a BLOB file will be stored as local file, nor bytes will be stored in class instance.
 * E.g. if we call AbstractValueData(InputStream, int, String) on BLOB we will store the stream content in a local file.
 *  
 * But for non-BLOB values constructor AbstractValueData(InputStream, int, String) 
 * will store data as array of bytes in memory.     
 * 
 * Created by The eXo Platform SARL        .
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: AbstractValueData.java 10024 2006-11-02 17:52:33Z peterit $
 */

public abstract class AbstractValueData implements ValueData {

  private static Log log = ExoLogger.getLogger("jcr.AbstractValueData");
  
  protected final InputStream stream;
  protected final byte[] data;

  protected int orderNumber; 

  /**
   * Create value data storing data as bytes array
   * 
   * @param data - bytes
   * @param orderNumber - value order nymber
   */
  protected AbstractValueData(final byte[] data, int orderNumber) {
    this.data = data;
    this.stream = null;
    
    this.orderNumber = orderNumber;
  }
  
  /**
   * Create value data storing data from input stream (given as external).
   * If 'inBuff' flag switched on then the stream will be stored as bytes array in memory.
   * Otherwise the stream will be used as direct source of data (with mark/reset abilities of a InputStream).
   * 
   * @param stream - InputStream instance 
   * @param orderNumber - value order nymber
   * @param inBuffer - if true, the stream will be stored as bytes array  
   */
  protected AbstractValueData(InputStream stream, int orderNumber, boolean inBuffer) throws IOException {
    
    if (inBuffer) {
      // read given stream into bytes buffer
      this.stream = null;
      this.data = BLOBUtil.readStream(stream);
      
      if (log.isDebugEnabled())
        log.debug(stream + " --> bytes. " + data.length + " bytes.");
    } else {
      // use given stream direct (will be used as data source), 
      // e.g. for CloneableInputStream streams, or user streams.
      this.data = null;
      this.stream = stream;
      
      if (log.isDebugEnabled())
        log.debug(stream + " --> stream.");
    }
    
    this.orderNumber = orderNumber;
  }

  /**
   * Create value data storing data as input stream (given as external) with swap option enabled.
   * I.e. if size of data stored in a stream more than maxBufferSize then this stream will be copied to swap file.
   * And value data will use this swap file instead use of original stream.
   * NOTE: stream will not stored in instance of ValueData.   
   * 
   * @param stream - InputStream instance 
   * @param orderNumber - value order nymber
   * @param swapStorage - swap file path (probably local or mounted on local file system)
   */
  protected AbstractValueData(InputStream stream, int orderNumber, BinaryValueSwapStorage swapStorage) throws IOException {    

    // TODO [PN] An idea: Make read cycle in independent thread with a result semaphore
    
    // Copy input stream to memory/swap except:
    // FileInputStream - use itself as data source (depend from swapStorage.isCopyLocalFiles() flag), will be cloned in getAsStream().
    // CloneableInputStream - use itself as data source as we think that the stream already a copy of existed stream.
    long fileStreamSize = 0;
    if (stream instanceof CloneableInputStream) {
      
      // For JCR internal streams, 
      // same logic as in constructor AbstractValueData(InputStream, int, boolean), with inBuffer=false.
      // Used from TransientValueData for incoming data streams.
      this.stream = stream;
      this.data = null;
      if (log.isDebugEnabled())
        log.debug(stream + " --> stream (cloneable)");

    } else if (stream instanceof FileInputStream && 
        (fileStreamSize = BLOBUtil.calcStreamLength(stream)) > swapStorage.getMinFileSize()) {
      
      if (swapStorage.isCopyLocalFiles()) {
        // copy the stream to swap/temp storage
        String swapFilePath = swapStorage.generatePath();
        
        TempFile swpFile = new TempFile(swapFilePath);
        
        FileOutputStream swpOut = new FileOutputStream(swpFile);
        fileStreamSize = BLOBUtil.copyFile((FileInputStream) stream, swpOut);
        swpOut.flush();
        swpOut.close();
        
        this.stream = new LazyTempFileInputStream(swpFile);
        this.data = null;
        
        //if (swapFilePath.indexOf("incoming")>0) {
        //  File ef = new File(swpFile.getAbsolutePath());
        //    log.info("TempFile: exists:" + ef.exists() + ", " + swapFilePath + ", " + fileStreamSize + "(" + ef.length() + ")" +" bytes.");
        //}
        
        if (log.isDebugEnabled())
          log.debug(stream + " --> stream (local copy). " + fileStreamSize + " bytes.");
        
      } else if (stream instanceof SharedFileInputStream) {
        // open lazy input stream to this stream file
        this.stream = ((SharedFileInputStream) stream).asLazy();
        this.data = null;
        if (log.isDebugEnabled())
          log.debug(stream + " --> shared stream. " + fileStreamSize + " bytes.");
      } else {
        // Use the stream direct as data source - no swap/temp storage, will be cloned each time in getAsStream().
        // it's a JCR users data usecase (e.g. Property.setValue(InputStream)).
        this.stream = stream;
        this.data = null;
        if (log.isDebugEnabled())
          log.debug(stream + " --> stream. " + fileStreamSize + " bytes.");        
      }
    } else {
      // Reading stream to bytes array in memory or to a swap file
      long start = System.currentTimeMillis();
      
      TempFile swpFile = null;
      FileChannel output = null;
      FileOutputStream outFile = null;
      
      ByteArrayOutputStream buff = null;
      String swapFilePath = null;
      
      try {          
        ReadableByteChannel input = Channels.newChannel(stream);
        
        ByteBuffer storageBuffer = null;
        int available = stream.available() + 64;
        int buffSize = BLOBUtil.calcBufferSize(available);
        if (available >= BLOBUtil.BLOB_MIN_AVAILABLE)
          storageBuffer = ByteBuffer.allocateDirect(buffSize);
        else
          storageBuffer = ByteBuffer.allocate(buffSize);
          
        int fileSize = 0; 
        int read = 0;
        while ((read = input.read(storageBuffer)) >= 0 || (read < 0 && !storageBuffer.hasRemaining())) {
          
          // Set the limit to the current position and the position to 0
          // making the new bytes visible for write()
          storageBuffer.flip();

          if (output != null) {
            // Write the bytes to the channel
            int numWritten = output.write(storageBuffer);
            fileSize += numWritten; //fileSize += read;
          } else if (buff == null && read > swapStorage.getMinFileSize()) {  
            // Threshold for keeping data in memory exceeded;
            // create swap file and spool buffer contents
            if (swapFilePath == null)
              swapFilePath = swapStorage.generatePath();
            else
              throw new IOException("Swap file name already assigned: " + swapFilePath);
            
            if (log.isDebugEnabled())
              log.debug("Writing stream " + stream + " to swap file..." + swapFilePath);
            
            swpFile = new TempFile(swapFilePath);
            outFile = new FileOutputStream(swpFile);
            output = outFile.getChannel();
            
            int numWritten = output.write(storageBuffer);
            fileSize += numWritten;
          } else if (buff != null && fileSize > swapStorage.getMinFileSize()) {
            // Threshold for keeping data in memory exceeded (after multiple reads);
            // create swap file and spool buffer contents
            if (swapFilePath == null)
              swapFilePath = swapStorage.generatePath();
            else
              throw new IOException("Swap file name already assigned: " + swapFilePath);
            
            if (log.isDebugEnabled())
              log.debug("Writing stream " + stream + " to swap file..." + swapFilePath);
            
            swpFile = new TempFile(swapFilePath);
            outFile = new FileOutputStream(swpFile);
            output = outFile.getChannel();
            
            int numWritten = output.write(ByteBuffer.wrap(buff.toByteArray())); // write from temp bytes output buffer
            numWritten = output.write(storageBuffer); // write current portion of data
            
            fileSize += numWritten; // was fileSize += read
          } else {
            // Storing data into temp bytes stream 
            if (buff == null)
              buff = new ByteArrayOutputStream();
            for (int i = 0; i < read; i++)
              buff.write(storageBuffer.get());
            fileSize += read;
          }
          
          // Check if all bytes were written
          if (storageBuffer.hasRemaining()) {
              // If not all bytes were written, move the unwritten bytes
              // to the beginning and set position just after the last
              // unwritten byte; also set limit to the capacity
            storageBuffer.compact();
          } else {
              // Set the position to 0 and the limit to capacity
            storageBuffer.clear();
          }
        }
      
        if (log.isDebugEnabled())
          log.debug(stream.getClass().getCanonicalName() + " --> "
            + (swpFile != null ? "swap file '" + swapFilePath + "'. " : "memory. ")
            + fileSize + " bytes. "
            + (System.currentTimeMillis() - start) + "ms.");
        
      } finally {
        if (output != null) {// channel
          output.close();
        }
        if (outFile != null) { // output stream
          outFile.flush();
          outFile.close();
        }       
      }
      
      if (swpFile != null) {
        // it's a stream based data (BLOB...)
        this.stream = new LazyTempFileInputStream(swpFile);
        this.data = null;
      } else {
        // it's a small data stored in memory (bytes...)
        // if no data readed store as empty array of bytes
        this.data = buff != null ? buff.toByteArray() : new byte[0];
        this.stream = null;          
      }
    }
  
    this.orderNumber = orderNumber;
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#getLength()
   */
  public long getLength() {
    try {
      return data != null ? data.length : BLOBUtil.calcStreamLength(stream);
    } catch(IOException e) {
      log.error("Can't calc data stream size: " + e.getMessage(), e);
      return -1;
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#getOrderNumber()
   */
  public final int getOrderNumber() {
    return orderNumber;
  }

  public final byte[] getAsByteArray() throws IllegalStateException {
    if(data != null) {
      return data;
    } else if(stream != null) {
      // [PN] An IllegalStateException will be throwed inside the call...
      return BLOBUtil.readStreamInmemory(stream);  
    } else {
      throw new NullPointerException("Data is null (as byte array)");
    }
  }
  
  public final boolean isByteArray() {
    return data != null;
  }

  public final InputStream getAsStream() throws IOException {
    // [PN] 07.09.06 We have to clone stream up from here
    if(stream != null) {
      if (stream instanceof CloneableInputStream)
        // clone stream using eXo JCR API
        return ((CloneableInputStream) stream).clone();
      else if (stream instanceof FileInputStream)
        // clone file stream using java.io.FileDescriptor 
        return new FileInputStream( ((FileInputStream) stream).getFD() );
      
      // Potentially danger operation as the ValueData will use the stream for copying self-data, 
      // but stream can be closed elsewhere outside.
      return stream;
    } else if(data != null)
      return new LazyByteArrayInputStream(data);
    else
      throw new NullPointerException("Data is null (as stream)");
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#setOrderNumber(int)
   */
  public void setOrderNumber(int orderNumber) {
    this.orderNumber = orderNumber;
  }
  
  protected void finalize() throws Throwable {
    try {
      try {
        if (stream != null)
          stream.close();
      } catch(IOException e) {
        // not matter 
      }
    } finally {
      super.finalize();
    }
  }
}
