/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl;

import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import javax.jcr.NamespaceRegistry;
import javax.jcr.RepositoryException;
import javax.jcr.nodetype.NodeTypeManager;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;

import org.exoplatform.container.ExoContainer;
import org.exoplatform.container.jmx.MX4JComponentAdapterFactory;
import org.exoplatform.services.jcr.access.AccessControlPolicy;
import org.exoplatform.services.jcr.config.BinarySwapEntry;
import org.exoplatform.services.jcr.config.DefaultBinarySwapEntry;
import org.exoplatform.services.jcr.config.RepositoryConfigurationException;
import org.exoplatform.services.jcr.config.RepositoryEntry;
import org.exoplatform.services.jcr.config.WorkspaceEntry;
import org.exoplatform.services.jcr.core.nodetype.ExtendedNodeTypeManager;
import org.exoplatform.services.jcr.impl.core.LocationFactory;
import org.exoplatform.services.jcr.impl.core.NamespaceDataPersister;
import org.exoplatform.services.jcr.impl.core.NamespaceRegistryImpl;
import org.exoplatform.services.jcr.impl.core.RepositoryImpl;
import org.exoplatform.services.jcr.impl.core.SessionFactory;
import org.exoplatform.services.jcr.impl.core.WorkspaceInitializer;
import org.exoplatform.services.jcr.impl.core.access.AccessManager;
import org.exoplatform.services.jcr.impl.core.lock.LockManager;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeDataPersister;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeManagerImpl;
import org.exoplatform.services.jcr.impl.core.observation.ObservationManagerRegistry;
import org.exoplatform.services.jcr.impl.core.query.QueryHandlerContext;
import org.exoplatform.services.jcr.impl.core.query.QueryManagerFactory;
import org.exoplatform.services.jcr.impl.core.query.SearchManager;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;
import org.exoplatform.services.jcr.impl.dataflow.persistent.BatcheableWorkspaceDataManager;
import org.exoplatform.services.jcr.impl.dataflow.persistent.WorkspaceStorageCacheImpl;
import org.exoplatform.services.jcr.impl.storage.value.StandaloneStoragePluginProvider;
import org.exoplatform.services.jcr.storage.swap.BinaryValueSwapStorage;
import org.exoplatform.services.jcr.storage.swap.BinaryValueTempStorage;
import org.exoplatform.services.jcr.util.UUIDGenerator;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: RepositoryContainer.java 10021 2006-11-02 10:27:25Z peterit $
 */

public class RepositoryContainer extends ExoContainer {

  private RepositoryEntry config;
  
  private MBeanServer mbeanServer;
  
  private BatcheableWorkspaceDataManager systemDataManager = null;


  public RepositoryContainer(ExoContainer parent, RepositoryEntry config)
  throws RepositoryException, RepositoryConfigurationException {

    //super(parent);
    super(new MX4JComponentAdapterFactory(), parent);

    // Defaults:
    if(config.getAccessControl() == null)
      config.setAccessControl(AccessControlPolicy.OPTIONAL);

    //this.name = config.getName();
    this.config = config;
    this.mbeanServer = MBeanServerFactory.createMBeanServer("jcrrep" + getName() + "mx");
    
    registerComponents();
    
  }

  /* (non-Javadoc)
   * @see org.picocontainer.defaults.DefaultPicoContainer#start()
   */
  @Override
  public void start() {
    
    try {
      
      RepositoryInitializer initializer = new RepositoryInitializer(this, config);
      initializer.init();
      
      initializer.load();
    } catch (RepositoryException e) {
      e.printStackTrace();
      throw new RuntimeException(e);
    } catch (RepositoryConfigurationException e) {
      e.printStackTrace();
      throw new RuntimeException(e);
    }
    
    super.start();
  }

  public MBeanServer getMBeanServer() {
    return this.mbeanServer;
  }

  /**
   * @return Returns the name.
   */
  public String getName() {
    return config.getName();
  }
  
  private void registerComponents() throws 
  RepositoryConfigurationException, RepositoryException {
    
    registerComponentInstance(config);

    registerWorkspacesComponents();
    registerRepositoryComponents();
  }
  
  
  private void registerRepositoryComponents() throws 
  RepositoryConfigurationException, RepositoryException {

    if (config.getBinaryTemp() != null) {
      BinaryValueTempStorage.configure(
          config.getBinaryTemp().getDirectoryPath(), 
          config.getBinaryTemp().getMinFileSize(),
          config.getBinaryTemp().getCopyLocalFiles());
    }
    
    registerComponentImplementation(UUIDGenerator.class);
    
    registerComponentImplementation(NamespaceDataPersister.class);
    registerComponentImplementation(NamespaceRegistryImpl.class);
    
    registerComponentImplementation(LocationFactory.class);
    registerComponentImplementation(ValueFactoryImpl.class);
    
    registerComponentImplementation(NodeTypeDataPersister.class);
    registerComponentImplementation(NodeTypeManagerImpl.class);
    
    registerComponentImplementation(AccessManager.class);
    try {
      registerComponentImplementation(Class.forName(config.getAuthenticationPolicy()));
    } catch (ClassNotFoundException e) {
      throw new RepositoryConfigurationException(
          "Class not found for repository authentication policy: " + e);
    }
    
    // Repository
    RepositoryImpl repository = new RepositoryImpl(this);
    registerComponentInstance(repository);

  }


  private void registerWorkspacesComponents() throws RepositoryException, RepositoryConfigurationException {
    List<WorkspaceEntry> wsEntries = config.getWorkspaceEntries();
    Collections.sort(wsEntries, new WorkspaceOrderComparator(config.getSystemWorkspaceName()));
    for (int i = 0; i < wsEntries.size(); i++) {
      registerWorkspace(wsEntries.get(i));      
    }
  }  
  
  private void registerWorkspace(WorkspaceEntry wsConfig) 
  throws RepositoryException, RepositoryConfigurationException {
  
    boolean isSystem = config.getSystemWorkspaceName().equals(wsConfig.getName());
    
    WorkspaceContainer workspaceContainer = new WorkspaceContainer(this, wsConfig);

    registerComponentInstance(wsConfig.getName(), workspaceContainer);
    
    // registerWorkspaceEnviroment(workspaceContainer, wsConfig);
    
    wsConfig.setUniqueName(getName() + "_" + wsConfig.getName());

    workspaceContainer.registerComponentInstance(wsConfig);
    
    // -------------- Binaty swap ----------------
    // [PN] TODO Back to swap as option, i.e. no default config must be found 
    if (wsConfig.getContainer().getBinarySwap() != null) {
      // using config provided by user 
      workspaceContainer.registerComponentInstance(wsConfig.getContainer().getBinarySwap());
    } else {
      // using default swap config
      workspaceContainer.registerComponentInstance(new DefaultBinarySwapEntry());
    }
    workspaceContainer.registerComponentImplementation(BinaryValueSwapStorage.class);
    
    workspaceContainer.registerComponentImplementation(StandaloneStoragePluginProvider.class);
    
    // re-registering one new value factory for this workspace only (BinaryValueSwapStorage will be used)
    workspaceContainer.registerComponentImplementation(ValueFactoryImpl.class);
    
    
    try {
      workspaceContainer.registerComponentImplementation(Class.forName(wsConfig.getContainer().getType()));
    } catch (ClassNotFoundException e) {
      throw new RepositoryConfigurationException(
          "Class not found for workspace data container " + wsConfig.getUniqueName() + ": " + e);
    }

    workspaceContainer.registerComponentImplementation(WorkspaceStorageCacheImpl.class);
    workspaceContainer.registerComponentImplementation(BatcheableWorkspaceDataManager.class);
    workspaceContainer.registerComponentImplementation(ObservationManagerRegistry.class);
    workspaceContainer.registerComponentImplementation(LockManager.class);
    
    if (wsConfig.getQueryHandler() != null) {
      try {
        workspaceContainer.registerComponentImplementation(QueryHandlerContext.class);
        Class qh = Class.forName(wsConfig.getQueryHandler().getType());
        workspaceContainer.registerComponentImplementation(qh);
      } catch (ClassNotFoundException e) {
        throw new RepositoryConfigurationException(
            "Class not found for workspace query handler, container " + wsConfig.getUniqueName() + ": " + e);
      }
      workspaceContainer.registerComponentImplementation(SearchManager.class);
      workspaceContainer.registerComponentImplementation(QueryManagerFactory.class);
    }
    workspaceContainer.registerComponentImplementation(WorkspaceInitializer.class);
    workspaceContainer.registerComponentImplementation(SessionFactory.class);
    
    BatcheableWorkspaceDataManager wsDataManager = 
      (BatcheableWorkspaceDataManager) workspaceContainer.getComponentInstanceOfType(
          BatcheableWorkspaceDataManager.class);
    if (isSystem) {
      // system workspace
      systemDataManager = wsDataManager;
      registerComponentInstance(systemDataManager);
    }
    wsDataManager.setSystemDataManager(systemDataManager);
  }
  
  // Components access methods -------

  public WorkspaceContainer getWorkspaceContainer(String workspaceName) {
    return (WorkspaceContainer)getComponentInstance(workspaceName);
  }

  public ExtendedNodeTypeManager getNodeTypeManager() {
    return (ExtendedNodeTypeManager)getComponentInstanceOfType(NodeTypeManager.class);
  }

  public NamespaceRegistry getNamespaceRegistry() {
    return (NamespaceRegistry) getComponentInstanceOfType(NamespaceRegistry.class);
  }

  public LocationFactory getLocationFactory() {
    return (LocationFactory)getComponentInstanceOfType(LocationFactory.class);
  }
  
  //  --------------------------------
  private static class WorkspaceOrderComparator implements Comparator<WorkspaceEntry> {
    private String sysWs;
    
    private WorkspaceOrderComparator(String sysWs) {
      this.sysWs = sysWs;
    }
    
    public int compare(WorkspaceEntry o1, WorkspaceEntry o2) {
      String n1 = o1.getName();
      return n1.equals(sysWs)?-1:0;
    }
  }

}