/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.Calendar;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.exoplatform.commons.utils.ISO8601;
import org.exoplatform.services.jcr.datamodel.ValueData;

/**
 * This class is the superclass of the type-specific
 * classes implementing the <code>Value</code> interfaces.
 *
 * @see javax.jcr.Value
 * @see org.exoplatform.services.jcr.impl.core.value.StringValue
 * @see org.exoplatform.services.jcr.impl.core.value.LongValue
 * @see org.exoplatform.services.jcr.impl.core.value.DoubleValue
 * @see org.exoplatform.services.jcr.impl.core.BoolanValue
 * @see org.exoplatform.services.jcr.impl.core.value.DateValue
 * @see org.exoplatform.services.jcr.impl.core.value.BinaryValue
 * @see org.exoplatform.services.jcr.impl.core.value.NameValue
 * @see org.exoplatform.services.jcr.impl.core.value.PathValue
 * @see org.exoplatform.services.jcr.impl.core.value.ReferenceValue
 *
 * @author Stefan Guggisberg
 */
public abstract class BaseValue implements Value {

  public static final String DEFAULT_ENCODING = "UTF-8";

  private static final short STATE_UNDEFINED = 0;

  private static final short STATE_VALUE_CONSUMED = 1;

  private static final short STATE_STREAM_CONSUMED = 2;

  private short state = STATE_UNDEFINED;

  protected final int type;

  protected InputStream stream = null;

  protected ValueData data;

  /**
   * Package-private default constructor.
   *
   * @param type The type of this value.
   */
  BaseValue(int type, ValueData data) {
    this.type = type;
    this.data = data;
  }

  /**
   * Checks if the non-stream value of this instance has already been
   * consumed (if any getter methods except <code>{@link #getStream()}</code> and
   * <code>{@link #getType()}</code> have been previously called at least once) and
   * sets the state to <code>STATE_STREAM_CONSUMED</code>.
   * @throws IllegalStateException if any getter methods other than
   *         <code>getStream()</code> and <code>getType()</code> have been
   *         previously called at least once.
   */
  protected void setStreamConsumed() throws IllegalStateException {

    if (state == STATE_VALUE_CONSUMED) {
      throw new IllegalStateException("non-stream value has already been consumed");
    }
    state = STATE_STREAM_CONSUMED;
  }

  /**
   * Checks if the stream value of this instance has already been
   * consumed (if {@link #getStream()} has been previously called
   * at least once) and sets the state to <code>STATE_VALUE_CONSUMED</code>.
   * @throws IllegalStateException if <code>getStream()</code> has been
   *         previously called at least once.
   */
  protected void setValueConsumed() throws IllegalStateException {
    if (state == STATE_STREAM_CONSUMED) {
      throw new IllegalStateException("stream value has already been consumed");
    }
    state = STATE_VALUE_CONSUMED;
  }

  /**
   * Returns the internal string representation of this value without modifying
   * the value state.
   * @return the internal string representation
   * @throws ValueFormatException if the value can not be represented as a
   * <code>String</code> or if the value is <code>null</code>.
   * @throws RepositoryException if another error occurs.
   */
  protected abstract String getInternalString() throws ValueFormatException, RepositoryException;

  //----------------------------------------------------------------< Value >
  /**
   * @see Value#getType
   */
  public int getType() {
    return type;
  }

  //----------------------------------------------------------------< Value >

  public Calendar getDate() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();
    Calendar cal = ISO8601.parse(getInternalString());
    if (cal == null) {
      throw new ValueFormatException("not a valid date format "+getInternalString());
    } else {
      return cal;
    }
  }

  public long getLong() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();
    try {
      return Long.parseLong(getInternalString());
    } catch (NumberFormatException e) {
      throw new ValueFormatException("conversion to long failed", e);
    }
  }

  public boolean getBoolean() throws ValueFormatException,
      IllegalStateException, RepositoryException {
    setValueConsumed();
    return Boolean.valueOf(getInternalString()).booleanValue();
  }

  public double getDouble() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();
    try {
      return Double.parseDouble(getInternalString());
    } catch (NumberFormatException e) {
      throw new ValueFormatException("conversion to double failed", e);
    }
  }

  public InputStream getStream() throws ValueFormatException,
      RepositoryException {
    setStreamConsumed();

    if (stream != null) {
      return stream;
    }

    try {
      // convert via string
      stream = new ByteArrayInputStream(getInternalString().getBytes(DEFAULT_ENCODING));
      return stream;
    } catch (UnsupportedEncodingException e) {
      throw new RepositoryException(DEFAULT_ENCODING + " not supported on this platform", e);
    }
  }

  public String getString() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();
    return getInternalString();
  }
  
  public String getReference() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    if(type != PropertyType.REFERENCE && type != PropertyType.BINARY)
      throw new ValueFormatException("Can not convert "+
          PropertyType.nameFromValue(type)+" to Reference");
    setValueConsumed();
    return getInternalString();
  }

  /**
   * @return Returns the data.
   */
  public ValueData getData() {
    return data;
  }
  
  /**
   * Returns the length of the value in bytes if the value is a PropertyType.BINARY, 
   * otherwise it returns the number of characters needed to display the value in its string form 
   * as defined in 6.2.6 Property Type Conversion. 
   * Returns –1 if the implementation cannot determine the length of the value..
   */
  public long getLength() {
    return data.getLength();
  }
}