/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;
import org.exoplatform.services.jcr.util.io.BLOBUtil;
import org.exoplatform.services.jcr.util.io.LazyByteArrayInputStream;
import org.exoplatform.services.log.ExoLogger;

/**
 * A <code>BinaryValue</code> provides an implementation of the
 * <code>Value</code> interface representing a binary value.
 *
 * @author Stefan Guggisberg
 */
public class BinaryValue extends BaseValue {

  public static final int TYPE = PropertyType.BINARY;

  protected static Log log = ExoLogger.getLogger("jcr.BinaryValue");

  /**
   * Constructs a <code>BinaryValue</code> object based on a string.
   *
   * @param text the string this <code>BinaryValue</code> should represent
   */
  public BinaryValue(String text) {
    super(TYPE, new TransientValueData(text));
  }

//  public BinaryValue(byte[] bytes) {
//    super(TYPE, new TransientValueData(bytes));
//  }

  /**
   * Constructs a <code>BinaryValue</code> object based on a bytes array.
   *
   * @param stream the stream this <code>BinaryValue</code> should represent
   */
  public BinaryValue(byte[] data) {
    super(TYPE, new TransientValueData(data));
  }
  
  /**
   * Constructs a <code>BinaryValue</code> object based on a stream.
   *
   * A stream value will be stored in workspace binary temporaty storage (aka swap) 
   * if its size is higher than BLOBUtil.minSwapFileSize.
   * Otherwise stream value will be stored in memory as bytes array. 
   *
   * The method will try to mark the stream with his size available for reading and read data as value,
   * then will try reset the stream for its marked position.
   *
   * @param stream the stream this <code>BinaryValue</code> should represent
   */
  public BinaryValue(InputStream stream) throws IOException {
    super(TYPE, new TransientValueData(stream));
  }

  /**
   * Constructs a <code>BinaryValue</code> object based on a stream.
   *
   * WARNING: A stream value will be used as value for created BinaryValue object. 
   * So, the stream given as parameter must be is not consumed. 
   * 
   * USE-CASE: If a stream of type BytesArray use this constructor with flag value 'true'.
   * That is a content of the stream will be copied to new binary value as bytes array.
   * Otherwise ('false' value) given stream will be stored inside TransientValueData object and will be used as data source.  
   * I.e. each one stream data, except of type BytesArray, will be shared in this value data object 
   * with other runtime readers of the stream. 
   * 
   * The method will try to mark the stream with his size available for reading and read data as value,
   * then will try reset the stream for its marked position.
   *
   * @param stream the stream this <code>BinaryValue</code> should represent
   */
  public BinaryValue(InputStream stream, boolean inBuffer) throws IOException {
    // TODO Use 'stream instanceof BytesArray' to decide flag value 
    super(TYPE, new TransientValueData(stream, inBuffer)); 
  }

  /**
   * Constructs a <code>BinaryValue</code> object based on a stream.
   *
   * @param data the stream this <code>BinaryValue</code> should represent
   */
  public BinaryValue(ValueData data) {
    super(TYPE, data);
  }

  /**
   * Indicates whether some other object is "equal to" this one.
   * <p>
   * The result is <code>true</code> if and only if the argument is not
   * <code>null</code> and is a <code>BinaryValue</code> object that
   * represents the same value as this object.
   *
   * @param obj the reference object with which to compare.
   * @return <code>true</code> if this object is the same as the obj argument;
   *         <code>false</code> otherwise.
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj instanceof BinaryValue) {
      BinaryValue other = (BinaryValue) obj;
      if (stream != null) {
        return stream.equals(other.stream);
      } else if (data != null) {
        return data.equals(other.getData());
      }
    }
    return false;
  }

  /**
   * Gets the string representaion of this binary value.
   */
  protected String getInternalString() throws ValueFormatException, RepositoryException {
    
    try {
      return new String(BLOBUtil.readValue(data), DEFAULT_ENCODING);
    } catch (UnsupportedEncodingException e) {
      throw new RepositoryException(DEFAULT_ENCODING + " not supported on this platform", e);
    } catch (IOException e) {
      throw new ValueFormatException("conversion from stream to string failed: " + e.getMessage(), e);
    } 
  }

  // ----------------------------------------------------------------< Value >
  /**
   * @see Value#getStream
   */

  public InputStream getStream() throws ValueFormatException, IllegalStateException, RepositoryException {
    
    setStreamConsumed();

    // build stream value if necessary
    if (stream == null) {
      if(data.isByteArray()) 
        stream = new LazyByteArrayInputStream(data.getAsByteArray());
      else
        try { 
          stream = data.getAsStream();
        } catch (IOException e) {
          throw new ValueFormatException("conversion from stream to stream failed: " + e.getMessage(), e);
        }
    }
    return super.getStream();
  }

}