/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.io.IOException;
import java.util.Calendar;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;
import org.exoplatform.services.jcr.util.io.BLOBUtil;
import org.exoplatform.services.log.ExoLogger;

/**
 * A <code>BooleanValue</code> provides an implementation of the
 * <code>Value</code> interface representing a boolean value.
 * 
 * @author Stefan Guggisberg
 */
public class BooleanValue extends BaseValue {

  public static final int TYPE = PropertyType.BOOLEAN;
  
  private static Log log = ExoLogger.getLogger("jcr.BooleanValue");

  private final Boolean   bool;

  /**
   * Constructs a <code>BooleanValue</code> object representing a boolean.
   * 
   * @param bool the boolean this <code>BooleanValue</code> should represent
   */
  public BooleanValue(boolean bool) {
    super(TYPE, new TransientValueData(bool));
    this.bool = new Boolean(bool);
  }

  BooleanValue(ValueData data) {
    super(TYPE, data);

    try {
      this.bool = new Boolean(new String(BLOBUtil.readValue(data)));
    } catch(IOException e) {
      throw new RuntimeException("FATAL ERROR IOException occured: " + e.getMessage(), e);
    }
  }

  /**
   * Indicates whether some other object is "equal to" this one. <p/> The result
   * is <code>true</code> if and only if the argument is not <code>null</code>
   * and is a <code>BooleanValue</code> object that represents the same value
   * as this object.
   * 
   * @param obj the reference object with which to compare.
   * @return <code>true</code> if this object is the same as the obj argument;
   *         <code>false</code> otherwise.
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj instanceof BooleanValue) {
      BooleanValue other = (BooleanValue) obj;
      if (bool == other.bool) {
        return true;
      } else if (bool != null && other.bool != null) {
        return bool.equals(other.bool);
      }
    }
    return false;
  }

  /**
   * @see BaseValue#getInternalString()
   */
  protected String getInternalString() throws ValueFormatException {
    if (bool != null) {
      return bool.toString();
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  // ----------------------------------------------------------------< Value >
  /**
   * @see Value#getDate
   */
  public Calendar getDate() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    throw new ValueFormatException("conversion to date failed: inconvertible types");
  }

  /**
   * @see Value#getLong
   */
  public long getLong() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    throw new ValueFormatException("conversion to long failed: inconvertible types");
  }

  /**
   * @see Value#getBoolean
   */
  public boolean getBoolean() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    if (bool != null) {
      return bool.booleanValue();
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  /**
   * @see Value#getDouble
   */
  public double getDouble() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    throw new ValueFormatException("conversion to double failed: inconvertible types");
  }
}
