/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.io.IOException;
import java.util.Calendar;
import java.util.Date;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;
import org.exoplatform.services.jcr.util.io.BLOBUtil;
import org.exoplatform.services.log.ExoLogger;

/**
 * A <code>DoubleValue</code> provides an implementation of the
 * <code>Value</code> interface representing a double value.
 * 
 * @author Stefan Guggisberg
 */
public class DoubleValue extends BaseValue {

  public static final int TYPE = PropertyType.DOUBLE;

  private static Log      log  = ExoLogger.getLogger("jcr.DoubleValue");

  private final Double    doubleNumber;

  /**
   * Constructs a <code>DoubleValue</code> object representing a double.
   * 
   * @param dbl the double this <code>DoubleValue</code> should represent
   */
  public DoubleValue(double dbl) {
    super(TYPE, new TransientValueData(dbl));
    this.doubleNumber = new Double(dbl);
  }

  DoubleValue(ValueData data) {
    super(TYPE, data);

    try {
      this.doubleNumber = new Double(new String(BLOBUtil.readValue(data)));
    } catch (IOException e) {
      throw new RuntimeException("FATAL ERROR IOException occured: " + e.getMessage(), e);
    }
  }

  /**
   * Indicates whether some other object is "equal to" this one. <p/> The result
   * is <code>true</code> if and only if the argument is not <code>null</code>
   * and is a <code>DoubleValue</code> object that represents the same value
   * as this object.
   * 
   * @param obj the reference object with which to compare.
   * @return <code>true</code> if this object is the same as the obj argument;
   *         <code>false</code> otherwise.
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj instanceof DoubleValue) {
      DoubleValue other = (DoubleValue) obj;
      if (doubleNumber == other.doubleNumber) {
        return true;
      } else if (doubleNumber != null && other.doubleNumber != null) {
        return doubleNumber.equals(other.doubleNumber);
      }
    }
    return false;
  }

  /**
   * @see BaseValue#getInternalString()
   */
  protected String getInternalString() throws ValueFormatException {
    if (doubleNumber != null) {
      return doubleNumber.toString();
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  // ----------------------------------------------------------------< Value >
  /**
   * @see Value#getDate
   */
  public Calendar getDate() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    if (doubleNumber != null) {
      // loosing timezone information...
      Calendar cal = Calendar.getInstance();
      cal.setTime(new Date(doubleNumber.longValue()));
      return cal;
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  /**
   * @see Value#getLong
   */
  public long getLong() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    if (doubleNumber != null) {
      return doubleNumber.longValue();
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  /**
   * @see Value#getBoolean
   */
  public boolean getBoolean() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    throw new ValueFormatException("conversion to boolean failed: inconvertible types");
  }

  /**
   * @see Value#getDouble
   */
  public double getDouble() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    if (doubleNumber != null) {
      return doubleNumber.doubleValue();
    } else {
      throw new ValueFormatException("empty value");
    }
  }
}
