/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.dataflow;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import javax.jcr.RepositoryException;

import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;

/**
 * Created by The eXo Platform SARL 
 * 
 * @author <a href="mailto:peter.nedonosko@exoplatform.com.ua">Peter Nedonosko</a>
 */
public class ItemDataFactory {

  public static ItemData makeItemData(ItemData itemData) throws RepositoryException {
    if (itemData.isNode())
      return makeNodeData((NodeData) itemData);
    
    return makePropertyData((PropertyData) itemData);
  }

  public static NodeData makeNodeData(NodeData nodeData) throws RepositoryException { 

    String uuid = nodeData.getUUID();
    if (uuid == null) {
      throw new RepositoryException("Node data has no UUID: "
          + (nodeData.getQPath() != null ? nodeData.getQPath().getAsString() : 
            "[node data has no path]"));
    }
    
    InternalQPath qpath = nodeData.getQPath();
    if (qpath == null) {
      throw new RepositoryException("Node data has no path: " 
          + nodeData.getUUID());
    }
    
    InternalQName ptn = nodeData.getPrimaryTypeName();
    
    // [PN] 17.08.06
    TransientNodeData transientData = new TransientNodeData(
        qpath, 
        new String(uuid), 
        nodeData.getPersistedVersion(),
        ptn, 
        nodeData.getMixinTypeNames(),
        nodeData.getOrderNumber(),
        nodeData.getParentUUID() != null ? new String(nodeData.getParentUUID()) : null,
        nodeData.getACL()
        );
    //transientData.setACL(nodeData.getACL());

    return transientData;
  }

  public static PropertyData makePropertyData(PropertyData propertyData) throws RepositoryException { 

    String uuid = propertyData.getUUID();
    if (uuid == null) {
      throw new RepositoryException("Property data has no UUID: "
          + (propertyData.getQPath() != null ? propertyData.getQPath().getAsString() : 
            "[property data has no path]"));
    }
    
    InternalQPath qpath = propertyData.getQPath();
    if (qpath == null) {
      throw new RepositoryException("Property data has no path: " 
          + propertyData.getUUID());
    }
    
    int pt = propertyData.getType();
    if (pt < 0) { //UNDEFINED=0
      throw new RepositoryException("Property data has no valid type: "  + pt + " "
          + propertyData.getQPath().getAsString() + " " + propertyData.getUUID());
    }

    List<ValueData> values = new ArrayList<ValueData>();
    for (ValueData ov : propertyData.getValues()) {
      values.add(makeValueData(ov));
    }    
    
    // [PN] 17.08.06
    TransientPropertyData transientData = new TransientPropertyData(
        qpath, 
        new String(uuid),
        propertyData.getPersistedVersion(),
        pt, 
        propertyData.getParentUUID() != null ? new String(propertyData.getParentUUID()) : null, 
        propertyData.isMultiValued()
        );
    
    transientData.setValues(values);
    return transientData;
  }

  public static ValueData makeValueData(ValueData valueData) throws RepositoryException {
    // [PN] 17.08.06 TransientValueData used
    int orderNumber = valueData.getOrderNumber();
    if (valueData.isByteArray()) {
      // make a copy of real data
      byte[] bytes = valueData.getAsByteArray();
      byte[] newBytes = new byte[bytes.length];  
      System.arraycopy(bytes, 0, newBytes, 0, newBytes.length);
      return new TransientValueData(newBytes, orderNumber);
    }
    try {
      return new TransientValueData(valueData.getAsStream(), orderNumber, false);
    } catch(IOException e) {
      throw new RepositoryException("ValueData copy error: " + e.getMessage(), e);
    }
  }
  
  public static List<NodeData> makeCopyOfNodes(List childNodes) throws RepositoryException {
    List<NodeData> copyOfChildsNodes = new LinkedList<NodeData>();
    synchronized (childNodes) {
      for (Object nodeData: childNodes) {
        if (nodeData instanceof NodeData) {
          NodeData copyOfItem = makeNodeData( (NodeData) nodeData);
          if (copyOfItem != null) {
            copyOfChildsNodes.add(copyOfItem);
          }
        } // else ignore, can check it after by size comparing
      }
    }
    return copyOfChildsNodes;
  }
  
  public static List<PropertyData> makeCopyOfProperties(List traverseProperties) throws RepositoryException {
    List<PropertyData> copyOfChildsProperties = new LinkedList<PropertyData>();
    synchronized (traverseProperties) {
      for (Object nodeProperty: traverseProperties) {
        if (nodeProperty instanceof PropertyData) {
          PropertyData copyOfItem = makePropertyData( (PropertyData) nodeProperty);
          if (copyOfItem != null) {
            copyOfChildsProperties.add(copyOfItem);
          } 
        } // else ignore, can check it after by size comparing
      }
    }
    return copyOfChildsProperties;
  }  
}
