/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.dataflow.persistent;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.WeakHashMap;

import org.apache.commons.logging.Log;
import org.exoplatform.services.cache.CacheService;
import org.exoplatform.services.cache.CachedObjectSelector;
import org.exoplatform.services.cache.ExoCache;
import org.exoplatform.services.cache.ObjectCacheInfo;
import org.exoplatform.services.jcr.config.CacheEntry;
import org.exoplatform.services.jcr.config.WorkspaceEntry;
import org.exoplatform.services.jcr.dataflow.persistent.WorkspaceStorageCache;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.ItemChangeState;
import org.exoplatform.services.jcr.impl.core.ItemState;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.dataflow.ItemsPersistenceListener;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL .
 * 
 * Info: This cache implementation store item data and childs lists of item data.
 *       And it implements OBJECTS cache - i.e. returns same java object that was cached before.
 *       Same item data or list of childs will be returned from getXXX() calls. [PN] 13.04.06
 * 
 * TODO: [PN] 18.04.06 To develop rules for synchronization once again with concurent testing
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov
 *         </a>
 * @version $Id: WorkspaceStorageCacheImpl.java 9668 2006-10-22 22:24:07Z peterit $
 */

public class WorkspaceStorageCacheImpl implements WorkspaceStorageCache, ItemsPersistenceListener {

  static public int MAX_CACHE_SIZE = 200;
  static public long MAX_CACHE_LIVETIME = 600; // in sec
  
  protected static Log log = ExoLogger.getLogger("jcr.WorkspaceStorageCache");

  private ExoCache cache; 
  // for each uuid in cache(C) we will store tree path->list-of-uuids for same
  private WeakHashMap<String, List<String>> pathCache;
  
  private WeakHashMap<String, List<NodeData>> nodesCache;
  private WeakHashMap<String, List<PropertyData>> propertiesCache;

  private String   name;

  private boolean  enabled;
  
  protected class ByPathRemoveSelector implements CachedObjectSelector {

    private final String parentPath;
    
    protected ByPathRemoveSelector(String parentPath) {
      this.parentPath = parentPath; 
    }
    
    public void onSelect(ExoCache exoCache, Serializable key, ObjectCacheInfo value) throws Exception {
      try {
        ItemData removed = (ItemData) exoCache.remove(key);
        if (removed != null) {
          exoCache.remove(removed.getUUID());
        }
      } catch(Exception e) {
        log.error("ByPathRemoveSelector.onSelect() " + parentPath + " key: " + key, e);
      }
    }

    public boolean select(Serializable key, ObjectCacheInfo value) {
      return ((String) key).startsWith(parentPath);
    }
    
  }
  
  public WorkspaceStorageCacheImpl(CacheService cacheService, WorkspaceEntry wsConfig)
      throws Exception {

    name = "jcr." + wsConfig.getUniqueName();
    cache = cacheService.getCacheInstance(name); // (WorkspaceDataCache) 
    
    CacheEntry cacheConfig = wsConfig.getCache();
    if (cacheConfig != null) {
      enabled = cacheConfig.isEnabled();
      int maxSize = Integer.parseInt(cacheConfig.getParameterValue("maxSize"));
      cache.setMaxSize(maxSize);
      pathCache = new WeakHashMap<String, List<String>>(maxSize);
      nodesCache = new WeakHashMap<String, List<NodeData>>(maxSize);
      propertiesCache = new WeakHashMap<String, List<PropertyData>>(maxSize);
      long liveTime = Long.parseLong(cacheConfig.getParameterValue("liveTime"));
      cache.setLiveTime(liveTime);
    } else {
      cache.setMaxSize(MAX_CACHE_SIZE);
      cache.setLiveTime(MAX_CACHE_LIVETIME);
      pathCache = new WeakHashMap<String, List<String>>();
      nodesCache = new WeakHashMap<String, List<NodeData>>();
      propertiesCache = new WeakHashMap<String, List<PropertyData>>();
      enabled = true;
    }
  }

  /**
   * @param uuid a UUID of item cached 
   * */
  public ItemData get(final String uuid) {
    if (!enabled)
      return null;
    
    try {
      return getItem(uuid);
    } catch (Exception e) {
      return null;
    }
  }
  
  /**
   * @param name
   * @return
   * @throws Exception
   */
  public ItemData get(final InternalQPath path) {
    if (!enabled)
      return null;
    
    try {
      return getItem(path);
    } catch (Exception e) {
      return null;
    }
  }  
  
  /**
   * Called by read operations
   * 
   * @param data
   */
  public void put(final ItemData data) {

    try {
      if (enabled && data != null) {
        
        // ExoCache calls synchronized already
        //synchronized (cache) {
          putItem(data);
        //}

        // add child item data to list of childs of the parent
        final String parentUUID = data.getParentUUID();
        if (data.isNode()) {
          // add child node
          final List<NodeData> cachedParentChilds = nodesCache.get(parentUUID);
          if (cachedParentChilds != null) {
            synchronized (cachedParentChilds) {
              int index = cachedParentChilds.indexOf(data);
              NodeData nodeData = (NodeData) data;
              if (index >= 0) {
                cachedParentChilds.set(index, nodeData);
                if (log.isDebugEnabled())
                  log.debug("put()    update child node  " + nodeData.getUUID() + "  at index #" + index);
              } else {
                cachedParentChilds.add(nodeData);
                if (log.isDebugEnabled())
                  log.debug("put()    add child node  " + nodeData.getUUID());
              }
            }
          }
        } else {
          // add child property
          final List<PropertyData> cachedParentChilds = propertiesCache.get(parentUUID);
          if (cachedParentChilds != null) {
            synchronized (cachedParentChilds) {
              int index = cachedParentChilds.indexOf(data);
              if (index >= 0) {
                cachedParentChilds.set(index, (PropertyData) data);
                if (log.isDebugEnabled())
                  log.debug("put()    update child property  " + data.getUUID() + "  at index #" + index);
              } else {
                if (cachedParentChilds.add((PropertyData) data)) {
                  if (log.isDebugEnabled())
                    log.debug("put()    add child property  " + data.getUUID());
                }
              }
            }
          }
        }
      }
    } catch (Exception e) {
      log.error("Error put item data in cache: "
          + (data != null ? data.getQPath().getAsString() : "[null]"), e);
    }
  }
  
  public void addChildProperties(final NodeData parentData, final List<PropertyData> childItems) {
    if (enabled && parentData != null && childItems != null ) { 

      String logInfo = null;
      if (log.isDebugEnabled()) {
        logInfo = "parent:   " + parentData.getQPath().getAsString() + "    " + parentData.getUUID() 
          + " " + childItems.size();
        log.debug("addChildProperties() >>> " + logInfo);
      }
      
      final String parentUUID = parentData.getUUID();
      String operName = ""; // for debug/trace only
      try {
        // [PN] 13.04.06
        // remove parent (no childs)
        operName = "removing parent";
        removeDeep(parentData, false);
        
        operName = "caching parent";
        putItem(parentData); // put parent in cache
       
        synchronized (propertiesCache) {
          // removing prev list of child properties from cache C
          operName = "removing child properties";
          removeChildProperties(parentUUID);
          
          operName = "caching child properties list";
          propertiesCache.put(parentUUID, childItems); // put childs in cache CP
        }
        
        operName = "caching child properties";
        putItems(childItems); // put childs in cache C
      } catch (Exception e) {
        log.error("Error in addChildProperties() " + operName + ": parent "
            + (parentData != null ? parentData.getQPath().getAsString() : "[null]"), e);
      }
      if (log.isDebugEnabled())
        log.debug("addChildProperties() <<< " + logInfo);
    }        
  }
  
  public void addChildNodes(final NodeData parentData, final List<NodeData> childItems) {
    //  && childItems.size() > 0 - no childs it's a state too
    if (enabled && parentData != null && childItems != null ) { 

      String logInfo = null;
      if (log.isDebugEnabled()) {
        logInfo = "parent:   " + parentData.getQPath().getAsString() + "    " + parentData.getUUID() 
          + " " + childItems.size();
        log.debug("addChildNodes() >>> " + logInfo);
      }
      
      final String parentUUID = parentData.getUUID();
      String operName = ""; // for debug/trace only
      try {
        // [PN] 13.04.06
        // remove parent (no childs)
        operName = "removing parent";
        removeDeep(parentData, false);
        
        operName = "caching parent";
        putItem(parentData); // put parent in cache
       
        synchronized (nodesCache) {
          // removing prev list of child nodes from cache C
          operName = "removing child nodes";
          final List<NodeData> removedChildNodes = removeChildNodes(parentUUID, false);
          
          if (removedChildNodes != null && removedChildNodes.size()>0) {
            operName = "search for stale child nodes not contains in the new list of childs";
            final List<NodeData> forRemove = new ArrayList<NodeData>();
            for (NodeData removedChildNode: removedChildNodes) {
              // used Object.equals(Object o), e.g. by UUID of nodes
              if (!childItems.contains(removedChildNode)) { 
                // this child node has been removed from the list
                // we should remve it recursive in C, CN, CP
                forRemove.add(removedChildNode);
              }
            }
          
            if (forRemove.size()>0) {
              operName = "removing stale child nodes not contains in the new list of childs";
              // do remove of removed child nodes recursive in C, CN, CP
              // we need here locks on cache, nodesCache, propertiesCache 
              synchronized (propertiesCache) {
                for (NodeData removedChildNode: forRemove) {
                  removeDeep(removedChildNode, true);
                }
              }
            }
          }
          
          operName = "caching child nodes list";
          nodesCache.put(parentUUID, childItems); // put childs in cache CN
        }
        
        operName = "caching child nodes";
        putItems(childItems); // put childs in cache C
      } catch (Exception e) {
        log.error("Error in addChildNodes() " + operName + ": parent "
            + (parentData != null ? parentData.getQPath().getAsString() : "[null]"), e);
      }
      if (log.isDebugEnabled()) 
        log.debug("addChildNodes() <<< " + logInfo);
    }    
  }
  
  protected void putItem(final ItemData data) throws Exception {
    final String path = data.getQPath().getAsString();
    final String uuid = data.getUUID();
    
    if (log.isDebugEnabled())
      log.debug("putItem()    " + path + "    " + uuid + "  --  " + data);
    
    cache.put(path, data);
    cache.put(uuid, data);    
  }
  
  protected ItemData getItem(final String uuid) throws Exception {
  
    return (ItemData) cache.get(uuid);
  }
  
  /**
   * @param key a InternalQPath path of item cached
   */
  protected ItemData getItem(final InternalQPath path) throws Exception {

    // ask direct cache (C)
    return (ItemData) cache.get(path.getAsString());
  }  
  
  protected void putItems(final List itemsList) throws Exception {
    for (Object item: itemsList) {
      putItem((ItemData) item);
    }
  }
  
  public List<NodeData> getChildNodes(final NodeData parentData) {
    
    if (!enabled)
      return null;
    
    try {
      return nodesCache.get(parentData.getUUID());
    } catch (Exception e) {
      log.error("Error in getChildNodes() parentData: " + 
          (parentData != null ? parentData.getQPath().getAsString() : "[null]"), e);
    }
    return null; // nothing cached
  }  
  
  public List<PropertyData> getChildProperties(final NodeData parentData) {
    
    if (!enabled)
      return null;
    
    try {
      return propertiesCache.get(parentData.getUUID());      
    } catch (Exception e) {
      log.error("Error in getChildNodes() parentData: " + 
          (parentData != null ? parentData.getQPath().getAsString() : "[null]"), e);
    }
    return null; // nothing cached
  }
  
  /**
   * @deprecated 
   * */
  private void clear() {
    if (!enabled)
      return ;
    
    try {
      if (log.isDebugEnabled())
        log.debug("<<<<<<<<<<<<<<<<<<< clear() " + cache.getCacheSize() + " >>>>>>>>>>>>>>>>>>>>>>>>");
      cache.clearCache();
      nodesCache.clear();
      propertiesCache.clear();
      //log.debug("Cache cleared >> " + cache.getCacheSize());
    } catch (Exception e) {
      log.error("Error clear workspace cache", e);
      //throw new RepositoryException("Exception in clear Cache " + e);
    }
  }

  public boolean isEnabled() {
    return enabled;
  }

  public void setEnabled(boolean enabled) {
    this.enabled = enabled;
  }

  public void setMaxSize(int maxSize) {
    this.cache.setMaxSize(maxSize);
  }

  public void setLiveTime(long liveTime) {
    this.cache.setLiveTime(liveTime);
  }
  
  // --------------------- ItemsPersistenceListener --------------
  public /*synchronized*/ void onBatchSaveItems(List itemStates, SessionImpl session) {
  }

  // [PN] 18.04.06 TODO to develop more soft synchronization behaviour 
  public synchronized void onSaveItems(final List itemStates, final SessionImpl session) {
    
    if (!enabled)
      return ;
    
    for (Iterator i = itemStates.iterator(); i.hasNext();) {
      ItemState state = (ItemState) i.next();
      ItemData data = state.getData();
      if (log.isDebugEnabled())
        log.debug("onSaveItems() " + ItemChangeState.getStateName(state.getState()) + " " + data.getQPath().getAsString() + " " + data.getParentUUID());

      try {
        if (state.getState() == ItemChangeState.ADDED || state.getState() == ItemChangeState.AUTO_ADDED 
            || state.getState() == ItemChangeState.UPDATED) {
          if (!data.isNode() && data.getQPath().getName().equals(Constants.JCR_MIXINTYPES)) {
            //if it's a property and it's jcr:mixinTypes we have to remove parent 
            //from cache, as it out of date
            ItemData parentData = get(data.getParentUUID());
            //remove(parentData);
            removeDeep(parentData, false);
          }
          put(data);
        } else if (state.getState() == ItemChangeState.DELETED || state.getState() == ItemChangeState.AUTO_DELETED) {
          if (!data.isNode() && data.getQPath().getName().equals(Constants.JCR_MIXINTYPES)) {
            //if it's a property and it's jcr:mixinTypes we have to remove parent
            //from cache, as it out of date
            ItemData parentData = get(data.getParentUUID());
            //remove(parentData);
            removeDeep(parentData, false);          
          } else
            remove(data);
        }
      } catch (Exception e) {
        log.error("Error process onSaveItems action for item data: " 
            + (data != null ? data.getQPath().getAsString() : "[null]"), e);
      }
    }
  }

  // ---------------------------------------------------

  /**
   * Called by delete
   * 
   * @param data
   */
  public void remove(final ItemData data) {
    if (!enabled)
      return;

    try {
      if (log.isDebugEnabled())
        log.debug("remove() " + data.getQPath().getAsString() + " " + data.getUUID());
      
      // do actual deep remove
      if (data.isNode()) {
        synchronized (propertiesCache) {
          synchronized (nodesCache) {
            removeDeep(data, true);
          }
        }
        removeSuccessors(data.getQPath().getAsString());
      } else {
        removeDeep(data, false);
      }
    } catch (Exception e) {
      log.error("Error remove item data from cache: " + (data != null ? data.getQPath().getAsString() : "[null]"), e);
    }
  }
  
  /**
   * Deep remove of an item in all caches (C, CN, CP).   
   * Outside must be sinchronyzed by cache(C).
   * If forceDeep=true then it must be sinchronyzed by cache(CN,CP) too.
   * 
   * @param item - ItemData of item removing
   * @param forceDeep - if true then childs will be removed too, 
   *        item's parent childs (nodes or properties) will be removed also.
   *        if false - no actual deep remove will be done, 
   *        the item only and theirs 'phantom by uuid' if exists.
   * */
  protected ItemData removeDeep(final ItemData item, final boolean forceDeep) throws Exception {
    final String myPath = item.getQPath().getAsString();
    if (log.isDebugEnabled())
      log.debug("removeDeep(" + forceDeep + ") >>> item " + myPath + " " + item.getUUID());
    
    if (forceDeep) {
      removeRelations(item);
    }
    
    cache.remove(item.getUUID());
    final ItemData itemData = (ItemData) cache.remove(myPath);
    if (itemData != null && !itemData.getUUID().equals(item.getUUID())) {
      // same path but diff uuid node
      removeDeep(itemData, forceDeep);
    }
    if (log.isDebugEnabled())
      log.debug("removeDeep(" + forceDeep + ") <<< item " + myPath + " " + item.getUUID());
    return itemData;
  }
  
  /**
   * Remove item relations in the cache(C,CN,CP) by UUID.
   * Relations for a node it's a child nodes, properties and item in node's parent childs list.
   * Relations for a property it's a item in node's parent childs list. 
   * */
  protected void removeRelations(final ItemData item) {
    // removing child item data from list of childs of the parent
    try {
      if (item.isNode()) {
        // removing childs of the node
        if (removeChildNodes(item.getUUID(), true) != null) {
          if (log.isDebugEnabled())
            log.debug("removeRelations() removeChildNodes() " + item.getUUID());
        }
        if (removeChildProperties(item.getUUID()) != null) {
          if (log.isDebugEnabled())
            log.debug("removeRelations() removeChildProperties() " + item.getUUID());
        }
        
        // removing child from the node's parent child nodes list
        if (removeChildNode(item.getParentUUID(), item.getUUID()) != null) {
          if (log.isDebugEnabled())
            log.debug("removeRelations() removeChildNode(parentUUID, childUUID) " 
              + item.getParentUUID() + " " + item.getUUID());
        }        
      } else {
        // removing child from the node's parent properties list
        if (removeChildProperty(item.getParentUUID(), item.getUUID()) != null) {
          if (log.isDebugEnabled())
            log.debug("removeRelations() removeChildProperty(parentUUID, childUUID) " 
              + item.getParentUUID() + " " + item.getUUID());
        }
      }
    } catch (Exception e) {
      log.error("Error in removeRelations() item: "
          + (item != null ? item.getQPath().getAsString() : "[null]"), e);
    }
  }

  protected List<NodeData> removeChildNodes(final String parentUUID, final boolean forceDeep) throws Exception {
    final List<NodeData> childNodes = nodesCache.remove(parentUUID);
    if (childNodes != null) {
      // we have child nodes
      for (NodeData cn: childNodes) {
        // [PN] 14.04.06 no deep here
        removeDeep(cn, forceDeep);
      }
    }
    return childNodes;
  }
  
  protected List<PropertyData> removeChildProperties(final String parentUUID) throws Exception {
    // [PN] 29.03.06 remove() instead get()
    final List<PropertyData> childProperties = propertiesCache.remove(parentUUID);
    if (childProperties != null) {
      // we have child properties
      for (PropertyData cp: childProperties) {
        // [PN] 13.04.06 
        removeDeep(cp, false);
      }
    }
    return childProperties;
  } 
    
  protected PropertyData removeChildProperty(final String parentUUID, final String childUUID) throws Exception {
    final List<PropertyData> childProperties = propertiesCache.get(parentUUID);
    if (childProperties != null) {
      synchronized (childProperties) { // [PN] 14.04.06 TODO Does we need lock here? 
        int removedIndex = -1;
        for (int i=0; i<childProperties.size(); i++) {
          PropertyData cn = childProperties.get(i);
          if (cn.getUUID().equals(childUUID)) {
            removedIndex = i;
            break;
          }
        }
        if (removedIndex>=0) {
          return childProperties.remove(removedIndex);
        }
      }
    }
    return null;
  }
  
  protected NodeData removeChildNode(final String parentUUID, final String childUUID) throws Exception {
    final List<NodeData> childNodes = nodesCache.get(parentUUID);
    if (childNodes != null) {
      synchronized (childNodes) { // [PN] 14.04.06 TODO Does we need lock here?
        int removedIndex = -1;
        for (int i=0; i<childNodes.size(); i++) {
          NodeData cn = childNodes.get(i);
          if (cn.getUUID().equals(childUUID)) {
            removedIndex = i;
            break;
          }
        }
        if (removedIndex>=0) {
          return childNodes.remove(removedIndex);
        }
      }
    }
    return null;
  }
  
  /** Not used */
  @Deprecated
  private NodeData getChildNode(final String parentUUID, final String childUUID) throws Exception {
    final List<NodeData> childNodes = nodesCache.get(parentUUID);
    if (childNodes != null) {
      for (NodeData cn: childNodes) {
        if (cn.getUUID().equals(childUUID)) {
          return cn;
        }
      }
    }
    return null;
  }
  
  /** Not used */
  @Deprecated
  private PropertyData getChildProperty(final String parentUUID, final String childUUID) throws Exception {
    final List<PropertyData> childProperties = propertiesCache.get(parentUUID);
    if (childProperties != null) {
      for (PropertyData cp: childProperties) {
        if (cp.getUUID().equals(childUUID)) {
          return cp;
        }
      }
    }
    return null;
  }
  
  /**
   * Remove successors by parent path.
   * Path is a string, if an item's path starts with it then the item will be removed 
   * */
  protected void removeSuccessors(final String parentPath) {
    // [PN] 23.10.06 Using synchronized mechanism of ExoCache
    final ByPathRemoveSelector remover = new ByPathRemoveSelector(parentPath);
    try {
      cache.select(remover);
    } catch(Exception e) {
      log.error("removeSuccessors() " + parentPath, e);
    }
//    if (cache instanceof HashMap) {
//      final Map<String, Object> cacheMap = (Map<String, Object>) cache;
//      final List<String> forRemove = new ArrayList<String>();
//      long keysCount = 0;
//      for (String key: cacheMap.keySet()) {
//        if (key.startsWith(parentPath)) {
//          // we should remove the entry
//          forRemove.add(key);
//        }
//        keysCount++;
//      }
//      for (String key: forRemove) {
//        try {
//          final ItemData removed = (ItemData) cache.remove(key);
//          if (removed != null) {
//            cache.remove(removed.getUUID());
//          }
//        } catch(Exception e) {
//          log.error("removeSuccessors() " + parentPath + " key: " + key, e);
//        }
//      }
//      if (forRemove.size()>0 && log.isDebugEnabled()) 
//        log.debug("removeSuccessors() STAT keys:" + keysCount + " removed: " + forRemove.size());
//    }
  }
}
