/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.storage.jdbc;

import java.sql.Connection;
import java.sql.SQLException;

import javax.jcr.RepositoryException;
import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.util.UUIDGenerator;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL
 * Author : Peter Nedonosko
 *          peter.nedonosko@exoplatform.com.ua
 * 29.08.2006
 * @version $Id: JDBCConnection.java 8506 2006-09-01 15:48:45Z peterit $
 */
abstract public class JDBCConnection extends DBConstants {
  
  protected static Log log = ExoLogger.getLogger("jcr.JDBCConnection");
  
  protected final Connection dbConnection;
  protected final DataSource dataSource;
  protected final String containerName;
  protected final int containerNameLength;
  protected final boolean multiDb;
  
  /**
   * Create per one datasource JDBCConnection instance.
   * 
   * @param dataSource
   * @param containerName
   * @param multiDb
   * @throws SQLException
   */
  public JDBCConnection(DataSource dataSource, String containerName, boolean multiDb)  throws SQLException {
    
    this.dbConnection = null;
    this.dataSource = dataSource;
    this.multiDb = multiDb;
    prepareQueries();
    prepareStatements();
    
    this.containerName = containerName; 
    this.containerNameLength = containerName.length();
  }
  
  /**
   * Create per one connection JDBCConnection instance. 
   * 
   * @param dbConnection
   * @param containerName
   * @param multiDb
   * @throws SQLException
   */
  public JDBCConnection(Connection dbConnection, String containerName, boolean multiDb)  throws SQLException {
    
    this.dbConnection = dbConnection;
    
    // Fix for Sybase jConnect JDBC driver bug. 
    // Which throws SQLException(JZ016: The AutoCommit option is already set to false)
    // if conn.setAutoCommit(false) called twise or more times with value 'false'.
    if (dbConnection.getAutoCommit()) {
      dbConnection.setAutoCommit(false);
    }
    
    this.dataSource = null;
    this.multiDb = multiDb;
    prepareQueries();
    prepareStatements();
    
    this.containerName = containerName; 
    this.containerNameLength = containerName.length();
  }
  
  /**
   * Used in Single Db Connection classes for UUID related queries 
   * @param uuid
   * @return
   */
  protected String id2cid(final String uuid) {
    if (!multiDb && uuid != null && uuid.length() == UUIDGenerator.UUID_LENGTH) {
        return containerName + uuid;
    }
    // uuid already prepared for single db
    return uuid;
  }
  
  /**
   * Used in loadXYZRecord methods for extract real UUID from container value.
   * @param uuid
   * @return
   */
  protected String cid2id(final String uuid) {
    if (!multiDb && uuid != null && uuid.length() > UUIDGenerator.UUID_LENGTH) {
      return uuid.substring(containerNameLength);
    }
    return uuid;
  }   
  
  /**
   * Prepared queries at start time
   * @throws SQLException
   */
  abstract protected void prepareQueries() throws SQLException;
  
  /**
   * Prepared statemens at statr time as it recommends by JDBC API doc
   * @throws SQLException
   */
  abstract protected void prepareStatements() throws SQLException;
  
  /**
   * Return JDBC connection obtained from initialized data source.
   * NOTE: Helper can obtain one new connection per each call of the method  or return one obtained once.
   *  
   * @throws SQLException
   */
  abstract public Connection getJdbcConnection() throws SQLException;
  
  /**
   * @throws IllegalStateException if connection is closed
   */
  protected void checkIfOpened() throws IllegalStateException {
    if(!isOpened())
      throw new IllegalStateException("Connection is closed");
  }
  
  /** Will be used as implementation of
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#isOpened()
   */
  public boolean isOpened() {
    try {
      return !getJdbcConnection().isClosed();
    } catch (SQLException e) {
      log.error(e);
      return false;
    }
  }

  /** Will be used as implementation of
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#rollback()
   */
  public final void rollback() throws IllegalStateException, RepositoryException {
    checkIfOpened();
    try {
      Connection con = getJdbcConnection();
      con.rollback();
      con.close();
      con = null;
    } catch (SQLException e) {
      throw new RepositoryException(e);
    }
  }
  
  /** Will be used as implementation of 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#commit()
   */
  public final void commit() throws IllegalStateException, RepositoryException {
    checkIfOpened();
    try {
      Connection con = getJdbcConnection();
      con.commit();
      con.close();
      con = null;
    } catch (SQLException e) {
      throw new RepositoryException(e);
    }
  }  
}
