/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.storage.jdbc;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedValueData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.storage.swap.BinaryValueSwapStorage;
import org.exoplatform.services.jcr.util.io.CloneableInputStream;
import org.exoplatform.services.jcr.util.io.LazyInputStream;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL Author : Peter Nedonosko
 * peter.nedonosko@exoplatform.com.ua 29.08.2006
 * 
 * @version $Id: JDBCIOHelper.java 10021 2006-11-02 10:27:25Z peterit $
 */
public class JDBCIOHelper extends JDBCConnection {

  public static final int LAZY_CACHE_SIZE = 1024 * 200; // 200K
  
  protected static final long CONNECTION_USE_TIMEOUT              = 1000 * 60 * 2; // 2 min - timeout of use after that the conection will be releases
  protected static final long CONNECTION_USE_COUNT              = 10; // max 10 times a connection can be used before will be released
  
  protected final String INSERT_VALUE_SINGLEDB              = "insert into JCR_SVALUE(DATA, ORDER_NUM, PROPERTY_ID) VALUES(?,?,?)";
  protected final String INSERT_VALUE_MULTIDB               = "insert into JCR_MVALUE(DATA, ORDER_NUM, PROPERTY_ID) VALUES(?,?,?)";

  protected final String DELETE_VALUE_SINGLEDB              = "delete from JCR_SVALUE where PROPERTY_ID=?";
  protected final String DELETE_VALUE_MULTIDB               = "delete from JCR_MVALUE where PROPERTY_ID=?";

  protected final String FIND_VALUESKEYS_BY_PROPERTYID_SINGLEDB = "select ORDER_NUM, PROPERTY_ID from JCR_SVALUE where PROPERTY_ID=? order by ORDER_NUM";
  protected final String FIND_VALUESKEYS_BY_PROPERTYID_MULTIDB  = "select ORDER_NUM, PROPERTY_ID from JCR_MVALUE where PROPERTY_ID=? order by ORDER_NUM";
  
  protected final String FIND_VALUEDATA_BY_PROPERTYID_OREDERNUMB_SINGLEDB = "select DATA from JCR_SVALUE where PROPERTY_ID=? and ORDER_NUM=?";
  protected final String FIND_VALUEDATA_BY_PROPERTYID_OREDERNUMB_MULTIDB = "select DATA from JCR_MVALUE where PROPERTY_ID=? and ORDER_NUM=?";

  protected static Log log = ExoLogger.getLogger("jcr.JDBCIOChannel");

  
  // --------- lazy connection stuff ------------
  private class JDBCIOHelperLazyDataSource {
    
    private final DataSource helperDataSource;
    private final String helperContainerName;
    
    protected JDBCIOHelperLazyDataSource(String containerName, DataSource dataSource) {
      this.helperContainerName = containerName;
      this.helperDataSource = dataSource;
    }

    protected String getHelperContainerName() {
      return helperContainerName;
    }

    protected DataSource get() {
      return helperDataSource;
    }
  }
  
  private class LazyJDBCInputStreamConnection {
    
    private Connection connection;
    private String streamContainerName;
    private long connectionTime = 0;
    private volatile long connectionUsed = 0;
    
    private volatile LazyJDBCInputStream consumeLock = null;
    
    private Map<LazyJDBCInputStream, Long> users = new WeakHashMap<LazyJDBCInputStream, Long>();
    
    protected LazyJDBCInputStreamConnection(String containerName, Connection connection) {
      this.streamContainerName = containerName;
      this.connection = connection;
      this.connectionTime = System.currentTimeMillis();
    }

    protected String getStreamContainerName() {
      return streamContainerName;
    }
    
    protected boolean isClosed() {  
      return connection == null;
    }
    
    protected void addUser(LazyJDBCInputStream userStream) {
      
      synchronized (users) {
        users.put(userStream, System.currentTimeMillis());
        connectionUsed++;
      }
    }
    
    protected Connection get() throws SQLException {
      
      if (!isClosed()) {
        consumeLock = null;
        //users.put(userStream, System.currentTimeMillis());
        return connection;
      }
      
      if (log.isDebugEnabled()) 
        log.debug("Helper connection is closed " + this + ", " + ", helper: " + JDBCIOHelper.this 
          + ", " + (connection != null ? connection.isClosed() : "[null]")  
          + ", " + (System.currentTimeMillis() - connectionTime) + ", " + users.size());
      throw new SQLException("Helper connection closed. Live time: " + (System.currentTimeMillis() - connectionTime) + ". " + this);
    } 
    
    protected void close(final LazyJDBCInputStream userStream) throws SQLException {

      if (consumeLock == null)
        consumeLock = userStream;
      
      try {
      
        synchronized (users) {
          
          users.remove(userStream);
          
          if (log.isDebugEnabled()) 
            log.debug("close() " + this + ", " + (connection != null ? connection.isClosed() : "[null]")  
              + ", " + (System.currentTimeMillis() - connectionTime) + ", " + users.size());
        
          if (connection != null && users.size() <= 0 && consumeLock == userStream) {            
            try {
              if (!connection.isClosed())
                connection.close();
            } finally {
              connection = null;
              if (log.isDebugEnabled()) 
                log.debug("close " + this);
            }
          }
        }
      
      } finally {
        if (consumeLock == userStream)
          consumeLock = null;
      }
    }
        
    /**
     * Tells that connection is consumed, i.e. can be used or no.
     */
    private boolean consumed() {
      return System.currentTimeMillis() - connectionTime >= CONNECTION_USE_TIMEOUT ||
        connectionUsed >= CONNECTION_USE_COUNT;
    }
    
    /**
     * Tells that connection is consumed, i.e. can't be used anymore.
     * And mark this connection as active request of consume.
     */
    protected boolean isConsumed(LazyJDBCInputStream lock) {
      try {

        long limit = 0;
        if (lock != null) {
          while (consumeLock != null && consumeLock != lock && limit < CONNECTION_USE_TIMEOUT) {
            try {
              Thread.sleep(limit=+ 20);
              if (log.isDebugEnabled()) 
                log.debug(">>>> Lock wait for " + consumeLock);
            } catch(Throwable e) {
              log.error("Lock wait error: " + e, e);
            }            
          }
          // in case of limit >= CONNECTION_USE_TIMEOUT - it's deadlock, conn is consumed sure          
          if (limit >= CONNECTION_USE_TIMEOUT) 
            return true;
        }
        
        final boolean consumed = isClosed() || consumed();
        if (!consumed) {
          consumeLock = lock; // lock for first get() call
        }
        
        if (log.isDebugEnabled()) 
          log.debug("isConsumed() " + this + ", isClosed: " + isClosed() + ", consumed: " + consumed + ", "
              + (System.currentTimeMillis() - connectionTime) + ", " + users.size());
        
        return consumed;
      } catch(Exception e) {
        log.error("isConsumed(), error: " + e.getMessage(), e);
        return true;
      }
    }
  }
  
  private static Map<JDBCIOHelper, JDBCIOHelperLazyDataSource> lazyDataSources = new WeakHashMap<JDBCIOHelper, JDBCIOHelperLazyDataSource>();
  private static Map<JDBCIOHelper, LazyJDBCInputStreamConnection> lazyConnections = new WeakHashMap<JDBCIOHelper, LazyJDBCInputStreamConnection>();
  
  protected final BinaryValueSwapStorage swapStorage;
  
  private class JDBCIOHelperInputStream extends LazyInputStream {
    
    private InputStream dataStream;
    private Statement streamStatement;
    
    JDBCIOHelperInputStream(InputStream dataStream, Statement streamStatement) {
      this.dataStream = dataStream;
      this.streamStatement = streamStatement;
    }

    @Override
    protected InputStream stream() throws IllegalStateException, IOException {
      return dataStream;
    }

    @Override
    public void close() throws IOException {
      super.close();
      try {
        streamStatement.close();
      } catch(SQLException e) {
        String mesg = "Error of close operation: " + e.getMessage();
        log.error(mesg, e);
        throw new IOException(mesg);
      }
    }
  }
  
  protected class LazyJDBCInputStream extends CloneableInputStream {
    
    private final String cid;
    private final int orderNumber;
    
    private LazyJDBCInputStreamConnection streamConnection;
    
    private InputStream dataStream;
    private Statement valueStatement;
    
    private ByteArrayOutputStream cache = new ByteArrayOutputStream();
    
    LazyJDBCInputStream(String cid, int orderNumber) {
      this.cid = cid;
      this.orderNumber = orderNumber;
    }
        
    // ------------------ InputStream --------------------
    // [PN] TODO back here in 1.2 and make stream itself cacheable 
//    @Override
//    public int read() throws IOException {
//      int b = super.read();
//      //cache.write(b);
//      //if (cache.size() > LAZY_CACHE_SIZE) {
//      //  cache.close();
//      //  cache = null;
//      //}
//      return b;
//    }
//
//    @Override
//    public int read(byte[] b, int off, int len) throws IOException {
//      int res = super.read(b, off, len);
//      //cache.write(b, off, len);
//      return res;
//    }
//
//    @Override
//    public int read(byte[] b) throws IOException {
//      int res = super.read(b);
//      //cache.write(b);
//      return res;
//    }
    // ---------------------------------------------------
    
    /**
     * Return - JDBC binary stream of DATA field. 
     * NOTE: a request statement is not closes. Actual close will be performed in 
     * LazyJDBCInputStream.close() method in client code.  
     * 
     * @param cid - property id of desired value
     * @param orderNumber - value order number (1 - for single valued values, more othervise) 
     * @return - JDBC binary stream of DATA field 
     * @throws SQLException
     * @throws IOException
     */
    private InputStream getValueDataStream() throws SQLException {

      streamConnection = openLazyConnection(JDBCIOHelper.this, this);
      
      PreparedStatement findValueData = streamConnection.get().prepareStatement(FIND_VALUE_BY_PROPERTYID_OREDERNUMB);
      
      findValueData.setString(1, cid);
      findValueData.setInt(2, orderNumber);
      final ResultSet vrs = findValueData.executeQuery();
      if (vrs.next()) {
        this.valueStatement = findValueData;
        return vrs.getBinaryStream(COLUMN_DATA);
      }
      findValueData.close();
      
      // if no records - empty stream will be returned
      return new ByteArrayInputStream(new byte[] {});
    }
    
    protected InputStream stream() throws IllegalStateException, IOException {
      if (dataStream == null) {
        try {
          dataStream = getValueDataStream();
        } catch(SQLException e) {
          String msg = "Error of data stream read operation (property: " 
            + cid2id(cid) + ", " + orderNumber + "), error:" + e.getMessage();
          log.error(msg, e);
          throw new IOException(msg);
        }
      }
      return dataStream;
    }
        
    @Override
    public void close() throws IOException {
      if (dataStream != null)
        super.close(); // close stream
      try {
        // close statement
        if (valueStatement != null)
          valueStatement.close();

        // close stream jdbc conn
        if (streamConnection != null) {
          //try {
          try {
            streamConnection.close(this);
          } catch(SQLException e) {
            log.error("Can't close lazy stream database connection (property: "
                + cid2id(cid) + ", " + orderNumber + "), error:" + e.getMessage(), e);
          }
          //} catch(SQLException e) {
          //  log.error("Can't check if is closed lazy stream database connection (property: "
          //          + cid2id(cid) + ", " + orderNumber + "), error:" + e.getMessage(), e);
          //}
        }
        
      } catch (SQLException e) {
        throw new IOException("Error of database statement close operation (property: "
            + cid2id(cid) + ", " + orderNumber + "), error:" + e);
      } finally {
        streamConnection = null;
      }
    }

    @Override
    public LazyJDBCInputStream clone() {
      return new LazyJDBCInputStream(cid, orderNumber);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj instanceof LazyJDBCInputStream) {
        LazyJDBCInputStream conStream = (LazyJDBCInputStream) obj;
        return conStream.cid.equals(cid) && conStream.orderNumber == orderNumber;
      }
      return super.equals(obj);
    }

    @Override
    public String toString() {
      return super.toString() + ", [" + JDBCIOHelper.this + ", " + cid2id(cid) + ", " + orderNumber + "]";
    }

    @Override
    protected void finalize() throws Throwable {
      
      this.close();
      
      super.finalize();
    }
    
  }
  
  JDBCIOHelper(Connection existedConn, DataSource lazyDataSource, String containerName, boolean multiDb, BinaryValueSwapStorage swapStorage) throws SQLException {
    super(existedConn, containerName, multiDb);
    this.swapStorage = swapStorage;

    JDBCIOHelper.lazyDataSources.put(this, new JDBCIOHelperLazyDataSource(containerName, lazyDataSource));
  }
  
  JDBCIOHelper(Connection existedConn, String containerName, boolean multiDb, BinaryValueSwapStorage swapStorage) throws SQLException {
    super(existedConn, containerName, multiDb);
    this.swapStorage = swapStorage;
  }
  
  @Override
  protected void finalize() throws Throwable {
    
    this.close();
    
    super.finalize();
  }

  @Override
  public String toString() {
    return super.toString() + ", [" + containerName + "]";
  }  
  
  @Override
  public Connection getJdbcConnection() throws SQLException { 
    if (dbConnection != null)
      return dbConnection;
    
    throw new SQLException("Helper connection is closed or released of use. It's possible helper is already consumed and closed");
  }
 
  private static synchronized LazyJDBCInputStreamConnection openLazyConnection(JDBCIOHelper helper, LazyJDBCInputStream userStream) throws SQLException {
    
    LazyJDBCInputStreamConnection streamConnection = null;
    
    while (streamConnection == null) {
    
      streamConnection = lazyConnections.get(helper);
      if (streamConnection == null || (streamConnection != null && streamConnection.isConsumed(userStream))) {
        
        // will open one new conn from helper data source
        JDBCIOHelperLazyDataSource lazyDataSource = lazyDataSources.get(helper);
        if (lazyDataSource != null) {
          Connection connection = lazyDataSource.get().getConnection();
          
          // using per-call transaction model
          if (!connection.getAutoCommit())
            connection.setAutoCommit(true);
          
          streamConnection = helper.new LazyJDBCInputStreamConnection(lazyDataSource.getHelperContainerName(), connection);
          lazyConnections.put(helper, streamConnection);
          
          if (log.isDebugEnabled()) 
            log.debug(streamConnection.getStreamContainerName() + ", open new jdbc connection, " 
                + ", " + streamConnection + ", closed:" + streamConnection.isClosed());
        } else {
          // no datasource is available - FATAL error, but impossible state due the constructor of a helper
          throw new SQLException("FATAL error! No data source is available for JDBCIOHelper instance " + helper);
        }
      } else {
        
//        if (streamConnection.isConsumed(userStream)) {
//          log.debug(streamConnection.getStreamContainerName() + ", skip jdbc connection, " 
//              + streamConnection + ", uses:" + streamConnection.users.size() + ", used:" + streamConnection.connectionUsed  
//              + ", in-use:" + (System.currentTimeMillis() - streamConnection.connectionTime) 
//              + "ms, closed:" + streamConnection.isClosed() );
//          streamConnection = null;
//        } else 
          
        if (log.isDebugEnabled()) { 
          log.debug(streamConnection.getStreamContainerName() + ", use jdbc connection, " 
              + streamConnection + ", uses:" + streamConnection.users.size() + ", used:" + streamConnection.connectionUsed  
              + ", in-use:" + (System.currentTimeMillis() - streamConnection.connectionTime) 
              + "ms, closed:" + streamConnection.isClosed() );
        }
      }
    }
    
    streamConnection.addUser(userStream);
    return streamConnection;
  }
 
  @Override
  protected void prepareQueries() throws SQLException {
    if (multiDb) {
      INSERT_VALUE = INSERT_VALUE_MULTIDB;
      DELETE_VALUE = DELETE_VALUE_MULTIDB;
      FIND_VALUES_BY_PROPERTYID = FIND_VALUESKEYS_BY_PROPERTYID_MULTIDB;
      FIND_VALUE_BY_PROPERTYID_OREDERNUMB = FIND_VALUEDATA_BY_PROPERTYID_OREDERNUMB_MULTIDB;
    } else {
      INSERT_VALUE = INSERT_VALUE_SINGLEDB;
      DELETE_VALUE = DELETE_VALUE_SINGLEDB;
      FIND_VALUES_BY_PROPERTYID = FIND_VALUESKEYS_BY_PROPERTYID_SINGLEDB;
      FIND_VALUE_BY_PROPERTYID_OREDERNUMB = FIND_VALUEDATA_BY_PROPERTYID_OREDERNUMB_SINGLEDB;
    }
  }

  /**
   * Nothing todo. Statements will be prepared before execution each time, as we
   * open connection by demand
   */
  @Override
  protected void prepareStatements() throws SQLException {
  }

  public void close() {
    
    // Do nothing as helper used once and will be cleaned by gc
    //dbConnection = null;
  }  
  
  public void delete(String propertyId) throws IOException {
    try {
      PreparedStatement deleteValue = getJdbcConnection().prepareStatement(DELETE_VALUE);
  
      try {
        deleteValue.setString(1, id2cid(propertyId));
        deleteValue.executeUpdate();
        getJdbcConnection().commit();
      } finally {
        deleteValue.close();
      }
    } catch (SQLException e) {
      try {
        getJdbcConnection().rollback();
      } catch (SQLException e1) {
        log.error("Error of delete rollback: " + e1.getMessage());
      }
      String msg = "Can't delete value data of property with id " + propertyId + ", error:" + e;
      log.error(msg, e);
      throw new IOException(msg);
    }
  }

  public void write(String propertyId, List<ValueData> data) throws IOException {
    try {
      String cid = id2cid(propertyId);
      for (int i = 0; i < data.size(); i++) {
        writeValueData(cid, data.get(i), i); // TODO data.get(i).getOrderNumber()
      }
      getJdbcConnection().commit();
    } catch (SQLException e) {
      try {
        getJdbcConnection().rollback();
      } catch (SQLException e1) {
        log.error("Error of write rollback: " + e1.getMessage());
      }
      String msg = "Can't write value data of property with id " + propertyId + ", error:" + e;
      log.error(msg, e);
      throw new IOException(msg);
    }
  }

  protected void writeValueData(String cid, ValueData data, int orderNumber) throws SQLException, IOException {

    InputStream stream = null;
    int streamLength = 0;
    if (data.isByteArray()) {
      byte[] dataBytes = data.getAsByteArray();
      stream = new ByteArrayInputStream(dataBytes);
      streamLength = dataBytes.length;
    } else {
      stream = data.getAsStream();
      streamLength = stream.available(); // for FileInputStream can be used channel.size() result
    }

    final PreparedStatement insertValue = getJdbcConnection().prepareStatement(INSERT_VALUE);

    try {
      insertValue.setBinaryStream(1, stream, streamLength);

      data.setOrderNumber(orderNumber);
      insertValue.setInt(2, data.getOrderNumber());

      insertValue.setString(3, cid);
      insertValue.executeUpdate();
    } finally {
      insertValue.close();
    }
  }

  public List<ValueData> read(String propertyId, boolean inBuffer) throws IOException {

    List<ValueData> data = new ArrayList<ValueData>();

    try {
      
      final PreparedStatement findValue = getJdbcConnection().prepareStatement(FIND_VALUES_BY_PROPERTYID);
  
      try {
        String cid = id2cid(propertyId);
        findValue.setString(1, cid);
        final ResultSet valueRecords = findValue.executeQuery();
        while (valueRecords.next()) {
          final int orderNum = valueRecords.getInt(COLUMN_ORDERNUM);
          ValueData vdata = null;
          if (swapStorage != null) {
            // open JDBC stream and read all data into BLOB swap 
            // (for small data in bytes buffer inside PersistedValueData contructor) 
            InputStream jdbcStream = readDataStream(cid, orderNum);
            try {
              vdata = new PersistedValueData(jdbcStream, orderNum, swapStorage);
            } finally {
              jdbcStream.close();
            }
          } else if (inBuffer) {
            // open JDBC stream and read all data into bytes buffer
            InputStream jdbcStream = readDataStream(cid, orderNum);
            try {
              vdata = new PersistedValueData(jdbcStream, orderNum, true);      
            } finally {
              jdbcStream.close();
            }
          } else {
            if (lazyDataSources.get(this) == null)
              log.warn("No data source available for LazyInputStream support now");
            
            // if no buffer/swap enabled - using lazy readed JDBC stream
            LazyJDBCInputStream lazyStream = new LazyJDBCInputStream(cid, orderNum);
            vdata = new PersistedValueData(lazyStream, orderNum);
          }
          data.add(vdata);
        }
      } finally {
        findValue.close();
      }
      
    } catch (SQLException e) {
      String msg = "Can't read value data of property with id " + propertyId + ", error:" + e;
      log.error(msg, e);
      throw new IOException(msg);
    }
    
    return data;
  }
  
  private InputStream readDataStream(String cid, int orderNumber) throws SQLException {
    
    final PreparedStatement findValueData = getJdbcConnection().prepareStatement(FIND_VALUE_BY_PROPERTYID_OREDERNUMB);

    findValueData.setString(1, cid);
    findValueData.setInt(2, orderNumber);
    final ResultSet vrs = findValueData.executeQuery();
    if (vrs.next()) {
      return new JDBCIOHelperInputStream(vrs.getBinaryStream(COLUMN_DATA), findValueData);
    }
    
    return new ByteArrayInputStream(new byte[] {});
  }
}
