/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.storage.jdbc;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedNodeData;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedPropertyData;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.storage.WorkspaceStorageConnection;
import org.exoplatform.services.jcr.storage.swap.BinaryValueSwapStorage;
import org.exoplatform.services.jcr.storage.value.ValueIOChannel;
import org.exoplatform.services.jcr.storage.value.ValueStoragePluginProvider;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL        .
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: JDBCStorageConnection.java 9984 2006-11-01 15:05:38Z peterit $
 */

abstract public class JDBCStorageConnection extends JDBCConnection implements WorkspaceStorageConnection {

  protected static Log log = ExoLogger.getLogger("jcr.JDBCStorageConnection");

  protected PreparedStatement findItemById;
  protected PreparedStatement findItemByPath;
  protected PreparedStatement findChildItemByPath;

  @Deprecated // [PN] TODO Refactor according JCR-56
  protected PreparedStatement findPropertiesByValue;

  protected PreparedStatement findReferenceable;
  protected PreparedStatement findReferences;

  protected PreparedStatement findValuesByPropertyId;

  protected PreparedStatement findNodesByParentId;
  protected PreparedStatement findNodesIdsByParentId;
  protected PreparedStatement findPropertiesByParentId;

  protected PreparedStatement findNodesCountByParentId;
  protected PreparedStatement findPropertiesCountByParentId;

  protected PreparedStatement insertItem;
  protected PreparedStatement insertNode;
  protected PreparedStatement insertProperty;
  protected PreparedStatement insertReference;

  protected PreparedStatement updateItem;
  protected PreparedStatement updateNode;
  protected PreparedStatement updateProperty;

  protected PreparedStatement deleteItem;
  protected PreparedStatement deleteNode;
  protected PreparedStatement deleteProperty;
  protected PreparedStatement deleteReference;

  protected final ValueStoragePluginProvider valueStorageProvider;
  protected final BinaryValueSwapStorage swapStorage;
  protected final DataSource dataSource;

  /**
   * @param dbConnection
   * @throws SQLException
   */
  protected JDBCStorageConnection(DataSource dataSource,
      String containerName, boolean multiDb,
      ValueStoragePluginProvider valueStorageProvider,
      BinaryValueSwapStorage swapStorage) throws SQLException {

    super(dataSource.getConnection(), containerName, multiDb);
    this.dataSource = dataSource;
    this.valueStorageProvider = valueStorageProvider;
    this.swapStorage = swapStorage;
    if (log.isDebugEnabled())
      log.debug("JDBCStorageConnection is opened using "+dbConnection.getClass().getName());
  }

  /**
   * Return JDBC connection obtained from initialized data source.
   * NOTE: Helper can obtain one new connection per each call of the method  or return one obtained once.
   *
   * @throws SQLException
   */
  public Connection getJdbcConnection() throws SQLException {
    return dbConnection;
  }

  // ---------------- SQLException handler -------------------

  protected String handleAddException(Exception e, ItemData item) throws RepositoryException, InvalidItemStateException {
    String message = "["+containerName+"] ";
    String errMessage = e.getMessage();
    String itemInfo = (item.isNode() ? "NODE " : "PROPERTY ")
      + item.getQPath().getAsString() + " " + item.getUUID()
      + (errMessage != null ? ". Cause: " + errMessage : "");

    if (errMessage != null)
      // try detect error by foreign key names
      if (errMessage.indexOf(JCR_FK_NODE_PARENT)>=0) {
        // we see that error contains JCR_FK_XNODE_PARENT constraint name, so it's constraint violation...
        message += "Parent not found for " + itemInfo;
        throw new InvalidItemStateException(message, e);
      } else if (errMessage.indexOf(JCR_FK_NODE_ITEM)>=0) {
        // Foreign key NODE->ITEM tables vioaltion. A item record not found for node created.
        message += "Item relation in JCR_ITEM table not found (not created). " + itemInfo;
        throw new RepositoryException(message, e);
      } else if (errMessage.indexOf(JCR_FK_PROPERTY_NODE)>=0) {
        // Foreign key PROPERTY->NODE tables vioaltion
        message += "Parent not found (not created) in JCR_NODE table. " + itemInfo;
        // [PN] 27.07.06 See JCR-170 7.1.1.6 Session.save()
        //throw new RepositoryException(message, e);
        throw new InvalidItemStateException(message, e);
      } else if (errMessage.indexOf(JCR_FK_PROPERTY_ITEM)>=0) {
        // Foreign key PROPERTY->ITEM tables vioaltion
        message += "Item relation in JCR_ITEM table not found (not created). " + itemInfo;
        throw new RepositoryException(message, e);
      } else if (errMessage.indexOf(JCR_FK_VALUE_PROPERTY)>=0) {
        // Foreign key VALUE->PROPERTY tables vioaltion
        message += "Property not found (not created) in JCR_PROPERTY table for values created. " + itemInfo;
        throw new RepositoryException(message, e);
      }

    // try detect integrity violation
    RepositoryException ownException = null;
    try {
      NodeData parent = (NodeData) getItemData(item.getParentUUID());
      if (parent != null) {
        // have a parent
        try {
          ItemData me = getItemData(item.getUUID());
          if (me != null) {
            // item already exists
            message += "Item already exists in storage: " + itemInfo;
            ownException = new RepositoryException(message, e);
            throw ownException;
          }
        } catch(Exception ep) {
          // item not found or other things but error of item reading
          if (ownException != null) throw ownException;
        }
        message += "Error add " + itemInfo;
        ownException = new RepositoryException(message, e);
        throw ownException;
      }
    } catch(Exception ep) {
      // no parent or error access it
      if (ownException != null) throw ownException;
    }
    message += "Parent not found (may be deleted by other session) for " + itemInfo;
    throw new InvalidItemStateException(message, e);
  }

  protected String handleDeleteException(Exception e, ItemData item) throws RepositoryException, InvalidItemStateException {
    String message = "["+containerName+"] ";
    String errMessage = e.getMessage();
    String itemInfo = (item.isNode() ? "NODE " : "PROPERTY ")
      + item.getQPath().getAsString() + " " + item.getUUID()
      + (errMessage != null ? ". Cause: " + errMessage : "");

    if (errMessage != null)
      // try detect error by foreign key names
      if (errMessage.indexOf(JCR_FK_NODE_PARENT)>=0) {
        // we see that error contains JCR_FK_MNODEPARENT constraint name, so it's constraint violation...
        message += "Can't delete parent till child exists " + itemInfo;
        throw new InvalidItemStateException(message, e);
      } else if (errMessage.indexOf(JCR_FK_NODE_ITEM)>=0) {
        // Foreign key NODE->ITEM tables vioaltion.
        message += "Can't delete item till related node exists in JCR_NODE table " + itemInfo;
        throw new RepositoryException(message, e);
      } else if (errMessage.indexOf(JCR_FK_PROPERTY_NODE)>=0) {
        // Foreign key PROPERTY->NODE tables vioaltion
        message += "Can't delete node till child property(ies) exists in JCR_PROPERTY table " + itemInfo;
        throw new RepositoryException(message, e);
      } else if (errMessage.indexOf(JCR_FK_PROPERTY_ITEM)>=0) {
        // Foreign key PROPERTY->ITEM tables vioaltion
        message += "Can't delete item till related property exists in JCR_PROPERTY table " + itemInfo;
        throw new RepositoryException(message, e);
      } else if (errMessage.indexOf(JCR_FK_VALUE_PROPERTY)>=0) {
        // Foreign key VALUE->PROPERTY tables vioaltion
        message += "Can't delete property till related property values exists in JCR_VALUE table " + itemInfo;
        throw new RepositoryException(message, e);
      }

    message += "Error of item delete " + itemInfo;
    throw new RepositoryException(message, e);
  }

  protected String handleUpdateException(Exception e, ItemData item) throws RepositoryException, InvalidItemStateException {
    String message = "["+containerName+"] ";
    String errMessage = e.getMessage();
    String itemInfo = (item.isNode() ? "NODE " : "PROPERTY ")
      + item.getQPath().getAsString() + " " + item.getUUID()
      + (errMessage != null ? ". Cause: " + errMessage : "");

    if (errMessage != null)
      // try detect error by foreign key names
      if (errMessage.indexOf(JCR_FK_VALUE_PROPERTY)>=0) {
        // Foreign key VALUE->PROPERTY tables vioaltion
        message += "Property not found (not created) in JCR_PROPERTY table for values updated. " + itemInfo;
        throw new RepositoryException(message, e);
      }

    // try detect integrity violation
    RepositoryException ownException = null;
    try {
      ItemData me = getItemData(item.getUUID());
      if (me != null) {
        // item already exists
        message += "Item exists but update error " + itemInfo;
        ownException = new RepositoryException(message, e);
        throw ownException;
      }
    } catch(Exception ep) {
      // item not found or other things but error of item reading
      if (ownException != null) throw ownException;
    }
    message += "Item not found (may be deleted by other session) " + itemInfo;
    throw new InvalidItemStateException(message, e);
  }

  // ---------------- WorkspaceStorageConnection -------------

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#add(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public void add(NodeData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      addNodeRecord(data);
    } catch (SQLException e) {
      handleAddException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#add(org.exoplatform.services.jcr.datamodel.PropertyData)
   */
  public void add(PropertyData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();

    try {
      addPropertyRecord(data);

      if (data.getType() == PropertyType.REFERENCE) {
        try {
          addReference(data);
        } catch(IOException e) {
          throw new RepositoryException("Can't read REFERENCE property ("+data.getQPath()+" "+data.getUUID()+") value: " + e.getMessage(), e);
        }
      }

      ValueIOChannel channel = valueStorageProvider.getApplicableChannel(data);
      if (channel != null) {
        channel.write(data.getUUID(), data.getValues());
        channel.close();
      } else {
        JDBCIOHelper helper = new JDBCIOHelper(dbConnection, containerName, multiDb, swapStorage);
        helper.write(data.getUUID(), data.getValues());
        helper.close();
      }

      if (log.isDebugEnabled())
        log.debug("values added "+data.getQPath()+" "+data.getUUID()+" use channel "+channel);
    } catch (IOException e) {
      handleAddException(e, data);
    } catch (SQLException e) {
      handleAddException(e, data);
    }
  }


  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#delete(org.exoplatform.services.jcr.datamodel.ItemData)
   */
  public void delete(ItemData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      if (data.isNode()) {
        if (deleteNode(id2cid(data.getUUID())) <= 0) {
          // [PN] it's error state, as we actually didn't delete the node (with sub-nodes)
          log.warn("Error state, the node (with sub-nodes) actually not deleted " + data.getQPath().getAsString());
        }
      } else {
        if (deleteProperty(id2cid(data.getUUID())) <= 0) {
          // [PN] it's error state, as we actually didn't delete the property
          log.warn("Error state, the property item actually not deleted " + data.getQPath().getAsString());
        }
      }
    } catch (SQLException e) {
      handleDeleteException(e, data);
    }
  }

  /**
   * @param cid, must be ready for container (e.g. if single-db then with prefix)
   */
  protected int deleteNode(String cid) throws SQLException, RepositoryException {
    int deleted = 0;

    // deleting child nodes
    ResultSet childNodes = findChildNodesIdsByParentUUID(cid);
    List<String> childNodesIds = new ArrayList<String>();
    while (childNodes.next()) {
      childNodesIds.add(childNodes.getString(COLUMN_ID));
    }
    childNodes.close();

    for (String childNodesId: childNodesIds) {
      int nc = deleteNode(childNodesId);
      if (nc <= 0) {
        // [PN] it's error state, as we actually didn't delete the child node
        log.warn("Error state, the node item actually not deleted " + childNodes.getString(COLUMN_PATH));
      }
      deleted += nc;
    }

    // deleting child properties
    ResultSet childProperties = findChildPropertiesByParentUUID(cid);
    while (childProperties.next()) {
      if (deleteProperty(childProperties.getString(COLUMN_PID)) <= 0) {
        // [PN] it's error state, as we actually didn't delete the property
        log.warn("Error state, the property item actually not deleted " + childProperties.getString(COLUMN_PATH));
      }
    }
    childProperties.close();

    // deleting from NODE table
    if (deleteNodeByUUID(cid) <= 0) {
      // [PN] it's error state, as we actually didn't delete the node
      log.warn("Error state, the node actually not deleted " + cid);
    }

    // deleting from ITEM table
    int nc = deleteItemByUUID(cid);
    if (nc <= 0) {
      // [PN] it's error state, as we actually didn't delete a item corresponding the node
      log.warn("Error state, a item corresponding the node actually not deleted " + cid);
    }
    deleted += nc;
    return deleted;
  }

  protected int deleteProperty(String cid) throws SQLException, RepositoryException {

    PropertyData data = (PropertyData) getItemByUUID(cid); // by container id

    try {
      // delete value
      ValueIOChannel channel = valueStorageProvider.getApplicableChannel(data);
      if (channel != null) {
        channel.delete(data.getUUID()); // by API UUI, not by cid
        channel.close();
      } else {
        JDBCIOHelper helper = new JDBCIOHelper(dbConnection, containerName, multiDb, swapStorage);
        helper.delete(data.getUUID());
        helper.close();
      }

      if (log.isDebugEnabled())
        log.debug("values deleted " + data.getUUID() + " use channel " + channel);

      deleteReference(cid);

      // delete property
      if (deletePropertyByUUID(cid) <= 0) {
        log.warn("Error state, the property actually is not deleted " + cid);
      }

      // delete item
      return deleteItemByUUID(cid);
    } catch (IOException e) {
      handleDeleteException(e, data);
      return 0; // will newer returns
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#update(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public void update(NodeData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      String cid = id2cid(data.getUUID());
      // version update
      updateItemVersionByUUID(data.getPersistedVersion(), cid);
      // order numb update
      updateNodeOrderNumbByUUID(data.getOrderNumber(), cid);
    } catch (SQLException e) {
      handleUpdateException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#update(org.exoplatform.services.jcr.datamodel.PropertyData)
   */
  public void update(PropertyData data) throws RepositoryException, UnsupportedOperationException,
  InvalidItemStateException, IllegalStateException {
    checkIfOpened();

    try {
      String cid = id2cid(data.getUUID());

      // version update
      updateItemVersionByUUID(data.getPersistedVersion(), cid);

      // update type
      updatePropertyTypeByUUID(data.getType(), cid);

      // update reference
      try {
        deleteReference(cid);

        if (data.getType() == PropertyType.REFERENCE) { //|| (data.getType() == PropertyType.UNDEFINED && data.getValues().size() <= 0)) {
          addReference(data);
        }
      } catch(IOException e) {
        throw new RepositoryException("Can't update REFERENCE property ("+data.getQPath()+" "+data.getUUID()+") value: " + e.getMessage(), e);
      }

      ValueIOChannel channel = valueStorageProvider.getApplicableChannel(data);
      if (channel != null) {
        channel.delete(data.getUUID()); // by API UUID, not by cid
        channel.write(data.getUUID(), data.getValues());
        channel.close();
      } else {
        JDBCIOHelper helper = new JDBCIOHelper(dbConnection, containerName, multiDb, swapStorage);
        helper.delete(data.getUUID()); // by API UUID, not by cid
        helper.write(data.getUUID(), data.getValues());
        helper.close();
      }
      
      if (log.isDebugEnabled())
        log.debug("values updated "+data.getQPath()+" "+data.getUUID()+" use channel "+channel);
    } catch (IOException e) {
      handleUpdateException(e, data);
    } catch (SQLException e) {
      handleUpdateException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildNodesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildNodesCount(NodeData nodeData) throws RepositoryException {
    checkIfOpened();
    try {
      return getChildNodesCountByParentUUID(id2cid(nodeData.getParentUUID())).getInt(1);
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildNodesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<NodeData> getChildNodesData(NodeData parent) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet node = findChildNodesByParentUUID(id2cid(parent.getUUID()));
      List<NodeData> childrens = new ArrayList<NodeData>();
      while(node.next()) {
        if (node.getString(COLUMN_NID) != null) {
          childrens.add(loadNodeRecord(node));
        } else {
          throw new RepositoryException("FATAL: Not found child node for parent "+parent.getQPath().getAsString()
              + ", but child item found " + node.getString(COLUMN_PATH) + " " + cid2id(node.getString(COLUMN_ID)));
        }
      }
      return childrens;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildPropertiesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildPropertiesCount(NodeData parent) throws RepositoryException {
    checkIfOpened();
    try {
      return getChildPropertiesCountByParentUUID(id2cid(parent.getUUID())).getInt(1);
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildPropertiesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<PropertyData> getChildPropertiesData(NodeData parent) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet prop = findChildPropertiesByParentUUID(id2cid(parent.getUUID()));
      List<PropertyData> children = new ArrayList<PropertyData>();
      while(prop.next()) {
        if (prop.getString(COLUMN_PID) != null) {
          children.add(loadPropertyRecord(prop));
        } else {
          throw new RepositoryException("FATAL: Not found child property for parent "+parent.getQPath().getAsString()
              + ", but child item found " + prop.getString(COLUMN_PATH) + " " + cid2id(prop.getString(COLUMN_ID)));
        }

      }
      return children;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } catch (IOException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getItemData(org.exoplatform.services.jcr.datamodel.InternalQPath)
   */
  public ItemData getItemData(InternalQPath qPath) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    ResultSet item = null;
    try {
      item = findItemByPath(qPath.getAsString());
      if(item.next()) {
        return itemData(item);
      }
      return null;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } catch (IOException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } finally {
      try {
        if (item != null)
          item.close();
      } catch(SQLException e) {
        log.error("getItemData() Error close resultset " + e.getMessage());
      }
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getItemData(java.lang.String)
   */
  public ItemData getItemData(String uuid) throws RepositoryException, IllegalStateException {
    return getItemByUUID(id2cid(uuid));
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getReferencesData(java.lang.String)
   */
  public List<PropertyData> getReferencesData(String nodeUUID) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet refProps = findReferences(id2cid(nodeUUID));
      List<PropertyData> references = new ArrayList<PropertyData>();
      while(refProps.next()) {
        references.add(loadPropertyRecord(refProps));
      }
      return references;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } catch (IOException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  // ------------------ Private methods ---------------

  protected ItemData getItemByUUID(String cid) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    ResultSet item = null;
    try {
      item = findItemByUUID(cid);
      if(item.next()) {
        return itemData(item);
      }
      return null;
    } catch (SQLException e) {
      throw new RepositoryException("getItemData() error", e);
    } catch (IOException e) {
      throw new RepositoryException("getItemData() error", e);
    } finally {
        try {
          if (item != null)
            item.close();
        } catch(SQLException e) {
          log.error("getItemData() Error close resultset " + e.getMessage());
        }
    }
  }

  /**
   * @param itemRecord
   * @return
   * @throws RepositoryException
   * @throws SQLException
   */
  private ItemData itemData(ResultSet itemRecord) throws RepositoryException, SQLException, IOException {
    String id = itemRecord.getString(COLUMN_ID);

    boolean isNode = itemRecord.getString(COLUMN_NID) != null;
    boolean isProperty = itemRecord.getString(COLUMN_PID) != null;

    if (isNode && isProperty) {
      // inconsistent data stored in db
      throw new RepositoryException("FATAL: Item stored as node and as property "+itemRecord.getString(COLUMN_PATH) + " " + cid2id(id));
    }

    if (!isNode && !isProperty) {
      // no data stored for item - node or property
      throw new RepositoryException("FATAL: Not found corresponding node nor property for item "+itemRecord.getString(COLUMN_PATH) + " " + cid2id(id));
    }

    if (isNode)
      return loadNodeRecord(itemRecord);

    // property
    return loadPropertyRecord(itemRecord);
    // found item record but no prop
    //throw new RepositoryException("FATAL: Not found corresponding property for item "+itemRecord.getString(COLUMN_PATH));
  }

  /**
   * Prepared statemens at statr time as it recommends by JDBC API doc
   * @throws SQLException
   */
  protected final void prepareStatements() throws SQLException {
    findItemById = dbConnection.prepareStatement(FIND_ITEM_BY_ID);
    findItemByPath = dbConnection.prepareStatement(FIND_ITEM_BY_PATH);
    findChildItemByPath = dbConnection.prepareStatement(FIND_CHILDITEM_BY_PATH);

    findPropertiesByValue = dbConnection.prepareStatement(FIND_PROPERTIES_BY_VALUE);

    findReferences = dbConnection.prepareStatement(FIND_REFERENCES);
    findReferenceable = dbConnection.prepareStatement(FIND_REFERENCEABLE);

    findValuesByPropertyId = dbConnection.prepareStatement(FIND_VALUES_BY_PROPERTYID);

    findNodesByParentId = dbConnection.prepareStatement(FIND_NODES_BY_PARENTID);
    findNodesIdsByParentId = dbConnection.prepareStatement(FIND_NODES_IDS_BY_PARENTID);
    findPropertiesByParentId = dbConnection.prepareStatement(FIND_PROPERTIES_BY_PARENTID);

    findNodesCountByParentId = dbConnection.prepareStatement(FIND_NODESCOUNT_BY_PARENTID);
    findPropertiesCountByParentId = dbConnection.prepareStatement(FIND_PROPERTIESCOUNT_BY_PARENTID);

    insertItem = dbConnection.prepareStatement(INSERT_ITEM);
    insertNode = dbConnection.prepareStatement(INSERT_NODE);
    insertProperty = dbConnection.prepareStatement(INSERT_PROPERTY);
    //insertValue = dbConnection.prepareStatement(INSERT_VALUE);
    insertReference = dbConnection.prepareStatement(INSERT_REF);

    updateItem = dbConnection.prepareStatement(UPDATE_ITEM);
    updateNode = dbConnection.prepareStatement(UPDATE_NODE);
    updateProperty = dbConnection.prepareStatement(UPDATE_PROPERTY);
    //updateValue = dbConnection.prepareStatement(UPDATE_VALUE);

    deleteItem = dbConnection.prepareStatement(DELETE_ITEM);
    deleteNode = dbConnection.prepareStatement(DELETE_NODE);
    deleteProperty = dbConnection.prepareStatement(DELETE_PROPERTY);
    //deleteValue = dbConnection.prepareStatement(DELETE_VALUE);
    deleteReference = dbConnection.prepareStatement(DELETE_REF);
  }

  protected PersistedNodeData loadNodeRecord(ResultSet item) throws RepositoryException, SQLException {

    String cNID = item.getString(COLUMN_NID);
    String cNPARENTID = item.getString(COLUMN_NPARENTID);
    int cVERSION = item.getInt(COLUMN_VERSION);
    int cNORDERNUM = item.getInt(COLUMN_NORDERNUM);
    InternalQPath qpath = InternalQPath.parse(item.getString(COLUMN_PATH));

    // PRIMARY
    InternalQPath ptPath = InternalQPath.makeChildPath(qpath, Constants.JCR_PRIMARYTYPE);

    ResultSet ptProp = findItemByPath(cNID, ptPath.getAsString());

    if (!ptProp.next())
      throw new RepositoryException("FATAL ERROR primary type record not found "
          + ptPath.getAsString()+", container "+this.containerName);

    ResultSet ptValue = findValuesByPropertyId(ptProp.getString(COLUMN_PID));

    if (!ptValue.next())
      throw new RepositoryException("FATAL ERROR primary type value not found "
          + ptPath.getAsString()+", container "+this.containerName);

    byte[] data = ptValue.getBytes(COLUMN_DATA);
    InternalQName ptName = InternalQName.parse(new String((data != null ? data : new byte[] {})));

    // MIXIN
    InternalQPath mtPath = InternalQPath.makeChildPath(qpath, Constants.JCR_MIXINTYPES);
    ResultSet mtProp = findItemByPath(cNID, mtPath.getAsString());

    InternalQName[] mixinNames = new InternalQName[0];
    if(mtProp.next()) {
      List<byte[]> mts = new ArrayList<byte[]>();
      ResultSet mtValues = findValuesByPropertyId(mtProp.getString(COLUMN_PID));
      while (mtValues.next()) {
        mts.add(mtValues.getBytes(COLUMN_DATA));
      }
      mixinNames = new InternalQName[mts.size()];
      for (int i = 0; i < mts.size(); i++) {
        mixinNames[i] = InternalQName.parse(new String(mts.get(i)));
      }
    }

    // ACL
    AccessControlList acl;
    if (isAccessControllable(mixinNames)) {

      InternalQPath ownerPath = InternalQPath.makeChildPath(qpath, Constants.NODE_OWNER_URI);

      PropertyData ownerData = (PropertyData) getItemData(ownerPath);

      InternalQPath permPath = InternalQPath.makeChildPath(qpath, Constants.NODE_PERM_URI);

      PropertyData permData = (PropertyData) getItemData(permPath);

      acl = new AccessControlList(ownerData, permData);
    } else {
      if(qpath.equals(Constants.ROOT_PATH)) // make default ACL for root
        acl = new AccessControlList(true);
      else
        acl = null;
    }

    return new PersistedNodeData(cid2id(cNID),
        qpath,
        cid2id(cNPARENTID),
        cVERSION,
        cNORDERNUM,
        ptName, mixinNames, acl);
  }


  private boolean isAccessControllable(InternalQName[] mixinNames) {
    for (int i = 0; i < mixinNames.length; i++) {
      if (mixinNames[i].getAsString().equals(Constants.ACCESS_TYPE_URI)) {
        return true;
      }
    }
    return false;
  }


  protected PersistedPropertyData loadPropertyRecord(ResultSet item) throws RepositoryException, SQLException, IOException {

    List<ValueData> values = new ArrayList<ValueData>();
    InternalQPath path = InternalQPath.parse(item.getString(COLUMN_PATH));
    String uuid = cid2id(item.getString(COLUMN_PID));
    PersistedPropertyData pdata = new PersistedPropertyData(uuid,
        path,
        cid2id(item.getString(COLUMN_PPARENTID)),
        item.getInt(COLUMN_VERSION),
        item.getInt(COLUMN_PTYPE),
        item.getBoolean(COLUMN_PMULTIVALUED)
    );

    ValueIOChannel channel = valueStorageProvider.getApplicableChannel(pdata);
    if (channel != null) {
      values = channel.read(uuid, pdata.getType() != PropertyType.BINARY);
      channel.close();
    } else {
      JDBCIOHelper helper = null;
      if (pdata.getType() != PropertyType.BINARY ||
        (pdata.getType() == PropertyType.BINARY && swapStorage != null))
        // Join to this JDBC connection if no binary value reading.
        // Join it also if it's a binary value but a swap is enabled (stream will be readed immediately to swap).
        helper = new JDBCIOHelper(dbConnection, containerName, multiDb, swapStorage);
      else
        // Otherwise will be used LazyInputStream with own opened on demand JDBC connection
        helper = new JDBCIOHelper(dbConnection, dataSource, containerName, multiDb, swapStorage);

      values = helper.read(uuid, pdata.getType() != PropertyType.BINARY);
      helper.close();
    }

    pdata.setValues(values);
    return pdata;
  }

  // ---- Data access methods (query wrappers) to override in concrete connection ------

  protected abstract void addNodeRecord(NodeData data) throws SQLException;

  protected abstract void addPropertyRecord(PropertyData prop) throws SQLException;

  protected abstract ResultSet findItemByPath(String path) throws SQLException;

  // find child item by path for parent identified by parentId
  protected abstract ResultSet findItemByPath(String parentId, String path) throws SQLException;

  protected abstract ResultSet findItemByUUID(String uuid) throws SQLException;

  // For PrimaryType value direct read from JDBC, no storage plugin use
  protected abstract ResultSet findValuesByPropertyId(String propertyId) throws SQLException;

  protected abstract void addReference(PropertyData data) throws SQLException, IOException;

  protected abstract void deleteReference(String propertyUuid) throws SQLException;

  protected abstract ResultSet findReferences(String nodeUuid) throws SQLException;

  protected abstract ResultSet findReferenceable(String propertyUuid) throws SQLException;

  @Deprecated
  protected abstract ResultSet findReferencesByValue(String uuid) throws SQLException;

  protected abstract ResultSet getChildNodesCountByParentUUID(String parentUUID) throws SQLException;

  protected abstract ResultSet getChildPropertiesCountByParentUUID(String parentUUID) throws SQLException;

  protected abstract ResultSet findChildNodesByParentUUID(String parentUUID) throws SQLException;

  // [PN] 22.05.06 For deleteNode() method
  protected abstract ResultSet findChildNodesIdsByParentUUID(String parentUUID) throws SQLException;

  protected abstract ResultSet findChildPropertiesByParentUUID(String parentUUID) throws SQLException;

  protected abstract int deleteItemByUUID(String uuid) throws SQLException;
  protected abstract int deleteNodeByUUID(String uuid) throws SQLException;
  protected abstract int deletePropertyByUUID(String uuid) throws SQLException;

  protected abstract int updateItemVersionByUUID(int versionValue, String uuid) throws SQLException;
  protected abstract int updateNodeOrderNumbByUUID(int orderNumb, String uuid) throws SQLException;
  protected abstract int updatePropertyTypeByUUID(int type, String uuid) throws SQLException;
}
