/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.storage.value.fs;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedValueData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.storage.swap.BinaryValueSwapStorage;
import org.exoplatform.services.jcr.storage.value.ValueIOChannel;
import org.exoplatform.services.jcr.util.io.BLOBUtil;
import org.exoplatform.services.jcr.util.io.LazyStorageFileInputStream;
import org.exoplatform.services.jcr.util.io.LazyTempFileInputStream;
import org.exoplatform.services.jcr.util.io.SharedFile;
import org.exoplatform.services.jcr.util.io.SharedFileInputStream;
import org.exoplatform.services.jcr.util.io.StorageFile;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL        .
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: FileIOChannel.java 8661 2006-09-07 10:17:32Z peterit $
 */

public class FileIOChannel implements ValueIOChannel {
  
  protected static Log log = ExoLogger.getLogger("jcr.FileIOChannel");
  
  protected final File rootDir;
  
  protected final BinaryValueSwapStorage swapStorage;
  
  public FileIOChannel(File rootDir, BinaryValueSwapStorage swapStorage) {
    this.rootDir = rootDir;
    this.swapStorage = swapStorage;
  }
  
  public FileIOChannel(File rootDir) {
    this(rootDir, null);
  }
  
  public List<ValueData> read(String propertyId, boolean inBuffer) throws IOException {
    
    final File[] valueFiles = rootDir.listFiles(new PropertyIDFilter(propertyId));
    if (valueFiles != null) {
      // Buffering is enabled for this property
      ArrayList<ValueData> data = new ArrayList<ValueData>(valueFiles.length);
      for (int orderNum = 0; orderNum<valueFiles.length; orderNum++) {
        StorageFile storageFile = StorageFile.create(valueFiles[orderNum].getAbsolutePath());
        ValueData vdata = null;
        if (swapStorage != null) {
          // Use the file as a BLOB swap file.
          // If swapStorage.isCopyLocalFiles() false then the file will not be copied 
          // and will be used as source for data direct.
          // Otherwise the file will be copied to a swap storage.
          SharedFileInputStream fis = new SharedFileInputStream(storageFile);
          try {
            vdata = new PersistedValueData(fis, orderNum, swapStorage);
          } finally {
            fis.close();
          }
        } else if (inBuffer) {
          // read the file content in a bytes buffer          
          FileInputStream fis = new FileInputStream(storageFile);
          try {
            vdata = new PersistedValueData(fis, orderNum, true); // read in bytes buffer
          } finally {
            fis.close();
          }
        } else {
          // use lazy readable stream
          LazyStorageFileInputStream fis = new LazyStorageFileInputStream(storageFile);
          vdata = new PersistedValueData(fis, orderNum);
        }
        data.add(vdata);
      }
      return data;
    }
    throw new IOException("READ. Can't list value files for property " + propertyId);
  }

  public void write(String propertyId, List<ValueData> values)  throws IOException {
    for(ValueData value: values) {
      long fileSize = -1;
      File file = null;
      FileOutputStream fileOutput = null;
      try {
        String fileName = propertyId + value.getOrderNumber();
        file = new File(rootDir, fileName);
        if (value.isByteArray()) {
          fileOutput = new FileOutputStream(file);
          byte[] buff = value.getAsByteArray();
          fileOutput.write(buff);
          fileSize = buff.length;
        } else {
          InputStream stream = null;
          try {
            stream = value.getAsStream();
            if (stream instanceof LazyTempFileInputStream) {
              LazyTempFileInputStream tempFileStream = (LazyTempFileInputStream) stream; 
              if (tempFileStream.moveTo(file)) {
                // if successful
                if (log.isDebugEnabled())
                  log.debug("Move value file to " + 
                    (file != null ? file.getAbsolutePath() : "[null]" ) + ". " + file.length() + " bytes.");
                return;
              } // else can't move, will copy content
              if (!tempFileStream.getSourceFile().exists() && file.exists()) {
                log.warn("Value data file created (size: " + file.length() 
                    + ") but content possible is not saved. Cause source file isn't exists " + tempFileStream.getSourceFile() 
                    + ". Try set copy-local-files parameter to 'false' on binary-swap configuration in workspace container (exo-jcr-config.xml)");
                return;
              }
            }
            fileOutput = new FileOutputStream(file);
            fileSize = BLOBUtil.writeFile(stream, fileOutput);
          } finally {
            if (stream != null)
              stream.close();
          }
        }
      } catch (FileNotFoundException e) {
        log.error("Write operation error: " + e.getMessage(), e);
      } finally {
        if (fileOutput != null)
          fileOutput.close();
      }
      if (log.isDebugEnabled())
        log.debug("Write file " + 
          (file != null ? file.getAbsolutePath() : "[null]" ) + ". " + fileSize + " bytes.");
    }
  }
  
  public void delete(String propertyId)  throws IOException {
    final File[] valueFiles = rootDir.listFiles(new PropertyIDFilter(propertyId));
    if (valueFiles != null)
      // remove old files
      for(File valueFile: valueFiles) {
        StorageFile storageFile = StorageFile.get(valueFile.getAbsolutePath());
        if (storageFile != null) {
          if (!storageFile.delete() && log.isDebugEnabled())
            log.debug("File can't be deleted now... will be deleted latter. " + valueFile.getAbsolutePath());
        } else { 
          log.warn("File in storage is out of storage map! Deleting as java.io.File. " + valueFile.getAbsolutePath());
          if (!valueFile.delete())
            log.warn("java.io.File can't be deleted. " + valueFile.getAbsolutePath());
        }
      }
    else 
      throw new IOException("DELETE. Can't list value files for property " + propertyId);
  }

  public void close() {
  }
  
  private class PropertyIDFilter implements FileFilter {
    
    private String id;

    public PropertyIDFilter(String id) {
      this.id = id;
    }

    public boolean accept(File file) {
      return file.getName().startsWith(id) && !file.getName().endsWith(SharedFile.DELETED_EXTENSION);
    }
  }
}
