/**
* Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
* Please look at license.txt in info directory for more license detail.   *
*/

package org.exoplatform.services.webdav.commands;

import java.util.ArrayList;
import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.Session;
import javax.jcr.version.Version;
import org.apache.commons.chain.Command;
import org.apache.commons.chain.Context;  
import org.apache.commons.logging.Log;
import org.exoplatform.services.log.ExoLogger;
import org.exoplatform.services.webdav.DavConst;
import org.exoplatform.services.webdav.WebDavCommandContext;
import org.exoplatform.services.webdav.props.PropManager;
import org.exoplatform.services.webdav.props.WebDavProperty;
import org.exoplatform.services.webdav.request.CommonPropDoc;
import org.exoplatform.services.webdav.request.WebDavRequest;
import org.exoplatform.services.webdav.response.DResponse;
import org.exoplatform.services.webdav.response.DavFileInfo;
import org.exoplatform.services.webdav.response.WebDavResponse;

/**
 * Created by The eXo Platform SARL
 * Author : Vitaly Guly <gavrik-vetal@ukr.net/mail.ru>
 * @version $Id: $
 */

public abstract class WebDavCommand implements Command {

  private static Log log = ExoLogger.getLogger("jcr.WebDavCommand");

  protected ThreadLocal<WebDavCommandContext> commandContext = new ThreadLocal<WebDavCommandContext>();
  
  protected boolean needSrcSession = true;
  protected boolean needDestSession = false;

  protected ThreadLocal<Session> srcSession = new ThreadLocal<Session>();
  protected ThreadLocal<Session> destSession = new ThreadLocal<Session>();
  
  public final boolean execute(Context context) throws Exception {
    commandContext.set((WebDavCommandContext)context);
    
    String defAuthHeader = davContext().getConfig().getAuthHeader();

    if (needSrcSession) {
      srcSession.set(davRequest().getSourceSession(davContext().getSessionProvider()));
      if (srcSession.get() == null) {
        log.info("Can't login to source workspace.");
        davResponse().answerAuthNeeded(defAuthHeader);        
        return false;
      }
    }
    
    if (needDestSession) {     
      if (!davRequest().isSameHosts()) {
        log.info("Source host and destination host are not equals.");        
        srcSession.get().logout();        
        davResponse().setResponseHeader(DavConst.Headers.WWWAUTHENTICATE, defAuthHeader);
        davResponse().answerForbidden();
        return false;
      }
      
      destSession.set(davRequest().getDestinationSession(davContext().getSessionProvider()));
      if (destSession.get() == null) {
        log.info("Can't login to destination workspace.");
        srcSession.get().logout();
        davResponse().answerAuthNeeded(defAuthHeader);
        return false;
      }
    }
    
    boolean status = process();

    if (srcSession.get() != null) {
      srcSession.get().logout();
    }
    if (destSession.get() != null) {
      destSession.get().logout();
    }
    
    return status;    
  }
    
  protected abstract boolean process();
  
  public final WebDavCommandContext davContext() {
    return commandContext.get();
  }
  
  public final WebDavRequest davRequest() {
    return commandContext.get().getWebDavRequest();
  }
  
  public final WebDavResponse davResponse() {
    return commandContext.get().getWebDavResponse();
  }
  
  public final Session jcrSrcSession() {
    return srcSession.get();
  }
  
  public final Session jcrDestSession() {
    return destSession.get();
  }
  
  protected DResponse getResponseByNode(Node node, String href, CommonPropDoc propsInfo) throws Exception {
    DResponse response = new DResponse();
    
    if (node instanceof Version) {
      response.setHref(href + DavConst.DAV_VERSIONPREFIX + node.getName());
    } else {
      response.setHref(href);
    }
    
    ArrayList<String> properties = propsInfo.getProperties();

    for (int i = 0; i < properties.size(); i++) {
      String propertyName = properties.get(i);
      WebDavProperty property = PropManager.getPropertyByName(propertyName);
      property.fillFromNode(node, href);
      response.AddProperty(property);
    }
    
    return response;
  }  
  
  protected DavFileInfo getResourceHeader(Node srcNode) throws Exception {
    DavFileInfo fileInfo = new DavFileInfo();
    
    fileInfo.setName(srcNode.getName());

    if (!srcNode.isNodeType(DavConst.NodeTypes.NT_FILE)) {
      if (srcNode.hasProperty(DavConst.NodeTypes.JCR_CREATED)) {
        fileInfo.setLatstModified(srcNode.getProperty(DavConst.NodeTypes.JCR_CREATED).getString());
      }      
      fileInfo.setContentType("");
      //fileInfo.setContentType("text/html;charset=UTF-8");
    } else {
      Node contentNode = srcNode.getNode(DavConst.NodeTypes.JCR_CONTENT);
      fileInfo.setLatstModified(contentNode.getProperty(DavConst.NodeTypes.JCR_LASTMODIFIED).getString());
      fileInfo.setContentType(contentNode.getProperty(DavConst.NodeTypes.JCR_MIMETYPE).getString());
      fileInfo.setInputStream(contentNode.getProperty(DavConst.NodeTypes.JCR_DATA).getStream());
      fileInfo.setResType(false);
    }
    
    return fileInfo;    
  }
  
  public DavFileInfo getResourceHeader() {
    try {
      Node srcNode = null;
      try {
        srcNode = (Node)jcrSrcSession().getItem(davRequest().getSrcPath()); 
      } catch (PathNotFoundException pexc) {
        return null;
      }
      return getResourceHeader(srcNode);
    } catch (Exception exc) {
      log.info("Can't get file header. " + exc.getMessage());
      exc.printStackTrace();
    }    
   return null;
  }
  
}
