/**
* Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
* Please look at license.txt in info directory for more license detail.   *
*/

package org.exoplatform.services.webdav.request;

import java.io.InputStream;
import javax.jcr.Session;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletRequestWrapper;
import org.apache.commons.logging.Log;
import org.exoplatform.services.log.ExoLogger;
import org.exoplatform.services.webdav.DavConst;
import org.exoplatform.services.webdav.DavUtil;
import org.exoplatform.services.webdav.WebDavSessionProvider;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * Created by The eXo Platform SARL
 * Author : Vitaly Guly <gavrik-vetal@ukr.net/mail.ru>
 * @version $Id: $
 */

public class WebDavRequestImpl extends HttpServletRequestWrapper implements WebDavRequest {

  private static Log log = ExoLogger.getLogger("jcr.WebDavRequestImpl");

  protected String serverPrefix;
  protected String serverApp;
  
  protected String sourcePath;
  protected String sourceWorkspace;
  protected String sourceVersion;
  
  protected String destinationPath;
  protected String destinationWorkspace;
  protected String destinationVersion;
  
  public WebDavRequestImpl(HttpServletRequest request) {
    super(request);
    
    serverPrefix = DavPathUtil.getServerPrefix(this);
    serverApp = DavPathUtil.getServletApp(this);
    
    sourceWorkspace = DavPathUtil.getSrcWorkspace(this);
    sourcePath = DavPathUtil.getSrcPath(this);
    
    if (request.getParameter(DavConst.DAV_VERSIONIDENTIFIER) != null) {
      sourceVersion = request.getParameter(DavConst.DAV_VERSIONIDENTIFIER);
    } else {
      if (sourcePath.indexOf("?") >= 0) {
        sourceVersion = sourcePath.substring(sourcePath.indexOf("?") + 1);
        sourceVersion = sourceVersion.substring(DavConst.DAV_VERSIONIDENTIFIER.length() + 1);
        sourcePath = sourcePath.substring(0, sourcePath.indexOf("?"));
      }      
    }
    
    if (getHeader(DavConst.Headers.DESTINATION) != null) {
      destinationWorkspace = DavPathUtil.getDestWorkspace(this);
      destinationPath = DavPathUtil.getDestPath(this);
    }
    
  }
  
  public String getLockToken() {
    String lockToken = getHeader(DavConst.Headers.LOCKTOKEN);
    if (lockToken != null) {      
      lockToken = lockToken.substring(1, lockToken.length() - 1);
      return lockToken;
    }

    String ifHeaderValue = getHeader(DavConst.Headers.IF);
    if (ifHeaderValue != null) {
      String headerLockToken = ifHeaderValue.substring(ifHeaderValue.indexOf("("));
      headerLockToken = headerLockToken.substring(2, headerLockToken.length() - 2);
      return headerLockToken;
      
    }
    
    return null;
  }

  protected void fillLockTokens(Session session) {
    String lockToken = getLockToken();
    if (lockToken != null) {
      session.addLockToken(lockToken);
    }
  }  
  
  public Session getSourceSession(WebDavSessionProvider sessionProvider) {    
    String authHeader = getHeader(DavConst.Headers.AUTHORIZATION);
    Session sourceSession = sessionProvider.getSession(authHeader, getSrcWorkspace());
    fillLockTokens(sourceSession);    
    return sourceSession;
  }
  
  public Session getDestinationSession(WebDavSessionProvider sessionProvider) {
    String authHeader = getHeader(DavConst.Headers.AUTHORIZATION);
    Session destinationSession = sessionProvider.getSession(authHeader, getDestWorkspace());
    fillLockTokens(destinationSession);
    return destinationSession;
  }
  
  public boolean isSameHosts() {
    return DavPathUtil.isSameDestHost(this, getServerPrefix());
  }
  
  public String getServerPrefix() {
    return serverPrefix;
  }
  
  public String getServerApp() {
    return serverApp;
  }
  
  public String getSrcWorkspace() {
    return sourceWorkspace;
  }
  
  public String getSrcPath() {
    return sourcePath;
  }
  
  public String getSrcVersion() {
    return sourceVersion;
  }
  
  public String getDestWorkspace() {
    return destinationWorkspace;
  }
  
  public String getDestPath() {
    return destinationPath;
  }
  
  public RequestDoc getDocumentFromRequest() {
    try {
      Document doc = DavUtil.GetDocumentFromRequest(this);
      
      if (doc == null) {        
        if ("PROPFIND".equals(this.getMethod())) {
          PropFindDoc propsDoc = new PropFindDoc();
          propsDoc.initWithDefaultProperties();
          return propsDoc;
        }
        return null;
      }
      
      Node documentNode = doc.getChildNodes().item(0);
      if (!DavConst.DAV_NAMESPACE.equals(documentNode.getNamespaceURI())) {
        log.info("Input document namespace not equals DAV:");
        return null;
      }
      
      String documentName = documentNode.getLocalName();
      RequestDoc requestDocument = null;

      for (int i = 0; i < DavConst.AVAILABLE_DOCUMENTS.length; i++) {
        if (documentName.equals(DavConst.AVAILABLE_DOCUMENTS[i][0])) {
          requestDocument = (RequestDoc)Class.forName(DavConst.AVAILABLE_DOCUMENTS[i][1]).newInstance();
        }
      }    
      
      if (requestDocument == null) {
        return null;
      }
      if (!requestDocument.initFromDocument(doc)) {
        return null;
      }
      return requestDocument;
    } catch (Exception exc) {
      log.info("Can't get document from request. " + exc.getMessage());
      exc.printStackTrace();
    }
    
    return null;
  }
  
  public InputStream getRequestStream() throws Exception {
    return getInputStream();
  }
  
  public String getContentType() {
    return getHeader(DavConst.Headers.CONTENTTYPE);
  }
  
}
